"""
Interface for reading EEA AqERep files (formerly known as Airbase data).
"""

import glob
import gzip
import logging
import os
import pathlib
import shutil
import tempfile

import numpy as np
import pandas as pd
from tqdm import tqdm

from pyaerocom.exceptions import EEAv2FileError, TemporalResolutionError
from pyaerocom.io.helpers import get_country_name_from_iso
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

from pyaerocom.units import Unit

logger = logging.getLogger(__name__)


class ReadEEAAQEREPBase(ReadUngriddedBase):
    """Class for reading EEA AQErep data

    Extended class derived from  low-level base class
    :class:`ReadUngriddedBase` that contains some more functionality.

    Note
    ----
    Currently only single variable reading into an :class:`UngriddedData`
    object is supported.
    """

    #: Mask for identifying datafiles
    _FILEMASK = "*.csv"

    #: Version log of this class (for caching)
    __version__ = "0.09"

    #: Column delimiter
    FILE_COL_DELIM = ","

    #: Name of the dataset (OBS_ID)
    DATA_ID = ""  # change this since we added more vars?

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: There is no global ts_type but it is specified in the data files...
    TS_TYPE = "variable"

    #: sampling frequencies found in data files
    TS_TYPES_FILE = {
        "hour": "hourly",
        "day": "daily",
    }

    #: Dictionary specifying values corresponding to invalid measurements
    #: there's no value for NaNs in this data set. It uses an empty string
    NAN_VAL = {}

    #: Dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    # There's only one variable in each file named concentration
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["concso2"] = "concentration"
    VAR_NAMES_FILE["conco3"] = "concentration"
    VAR_NAMES_FILE["concno2"] = "concentration"
    VAR_NAMES_FILE["concco"] = "concentration"
    VAR_NAMES_FILE["concno"] = "concentration"
    VAR_NAMES_FILE["concpm10"] = "concentration"
    VAR_NAMES_FILE["concpm25"] = "concentration"
    VAR_NAMES_FILE["vmro3"] = "concentration"
    VAR_NAMES_FILE["vmro3max"] = "concentration"
    VAR_NAMES_FILE["vmrno2"] = "concentration"

    VAR_NAMES_FILE["concSso2"] = "concentration"
    VAR_NAMES_FILE["concNno"] = "concentration"
    VAR_NAMES_FILE["concNno2"] = "concentration"

    #: units of variables in files (needs to be defined for each variable supported)
    VAR_UNITS_FILE = {
        "µg/m3": "ug m-3",
        "mg/m3": "mg m-3",
        "µgS/m3": "ug S m-3",
        "µgN/m3": "ug N m-3",
        "ppb": "ppb",
    }

    #: file masks for the data files
    FILE_MASKS = dict(
        concso2="**/??_1_*_timeseries.csv*",
        concpm10="**/??_5_*_timeseries.csv*",
        conco3="**/??_7_*_timeseries.csv*",
        vmro3="**/??_7_*_timeseries.csv*",
        vmro3max="**/??_7_*_timeseries.csv*",
        concno2="**/??_8_*_timeseries.csv*",
        vmrno2="**/??_8_*_timeseries.csv*",
        concco="**/??_10_*_timeseries.csv*",
        concno="**/??_38_*_timeseries.csv*",
        concpm25="**/??_6001_*_timeseries.csv*",
        concSso2="**/??_1_*_timeseries.csv*",
        concNno2="**/??_8_*_timeseries.csv*",
        concNno="**/??_38_*_timeseries.csv*",
    )

    # conversion factor between concX and vmrX
    CONV_FACTOR = {}

    CONV_FACTOR["concSso2"] = np.float64(0.50052292274792)
    CONV_FACTOR["concNno2"] = np.float64(0.3044517868011477)
    CONV_FACTOR["concNno"] = np.float64(0.466788868521913)
    CONV_FACTOR["vmro3"] = np.float64(
        0.493
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx
    CONV_FACTOR["vmro3max"] = np.float64(
        0.493
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx
    CONV_FACTOR["vmrno2"] = np.float64(
        0.514
    )  # retrieved using STD atmosphere from geonum and pya.mathutils.concx_to_vmrx

    # unit of the converted property after the conversion
    CONV_UNIT = {}
    CONV_UNIT["concSso2"] = "µgS/m3"
    CONV_UNIT["concNno2"] = "µgN/m3"
    CONV_UNIT["concNno"] = "µgN/m3"
    CONV_UNIT["vmro3"] = "ppb"
    CONV_UNIT["vmro3max"] = "ppb"
    CONV_UNIT["vmrno2"] = "ppb"

    #: field name of the start time of the measurement (in lower case)
    START_TIME_NAME = "datetimebegin"

    #: filed name of the end time of the measurement (in lower case)
    END_TIME_NAME = "datetimeend"

    #: dictionary that connects the EEA variable codes with aerocom variable names
    VAR_CODES = {}
    VAR_CODES["1"] = "concso2"
    VAR_CODES["5"] = "concpm10"
    VAR_CODES["7"] = "conco3"
    VAR_CODES["8"] = "concno2"
    VAR_CODES["10"] = "concco"
    VAR_CODES["38"] = "concno"
    VAR_CODES["6001"] = "concpm25"

    #: column name that holds the EEA variable code
    VAR_CODE_NAME = "airpollutantcode"

    #: List of variables that are provided by this dataset (will be extended
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    #: there's no general instrument name in the data
    INSTRUMENT_NAME = "unknown"

    #: max time steps to read per file (hourly data)
    # to make numpy array allocation size static
    MAX_LINES_TO_READ = 24 * 366

    #: file name of the metadata file
    #: this will be prepended with a data path later on
    # this file is in principe updated once a day.
    # so we night consider updating it from within the code later on.
    # URL: https://discomap.eea.europa.eu/map/fme/metadata/PanEuropean_metadata.csv
    DEFAULT_METADATA_FILE = "metadata.csv"

    #: Name of latitude variable in metadata file
    LATITUDENAME = "latitude"

    #: name of longitude variable in metadata file
    LONGITUDENAME = "longitude"

    #: name of altitude variable in metadata file
    ALTITUDENAME = "altitude"

    #: this class reads the European Environment Agency's Eionet data
    #: for details please read
    #: https://www.eea.europa.eu/about-us/countries-and-eionet
    WEBSITE = "https://discomap.eea.europa.eu/map/fme/AirQualityExport.htm"

    #: Eionet offers 2 data revisions
    #: E2a (near real time) and E1a (quality controlled)
    #: this class reads the E2a data for now.
    # But by changing the base path
    # and this constant, it can also read the E1a data set
    DATA_PRODUCT = ""

    AUX_REQUIRES = {
        "vmro3max": ["conco3"],
        "vmro3": ["conco3"],
        "vmrno2": ["concno2"],
        "concNno2": ["concno2"],
        "concNno": ["concno"],
        "concSso2": ["concso2"],
    }

    AUX_FUNS = {
        "vmro3": NotImplementedError(),
        "vmro3max": NotImplementedError(),
        "vmrno2": NotImplementedError(),
        "concNno2": NotImplementedError(),
        "concNno": NotImplementedError(),
        "concSso2": NotImplementedError(),
    }

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id=data_id, data_dir=data_dir)
        self._metadata = None

    @property
    def DEFAULT_VARS(self):
        """List of default variables"""
        return [self.VAR_CODES["7"]]

    @property
    def DATASET_NAME(self):
        """Name of the dataset"""
        return self.data_id

    def read_file(self, filename, var_name, vars_as_series=False):
        """Read a single EEA file

        Note that there's only a single variable in the file

        Parameters
        ----------
        filename : str
            Absolute path to filename to read.
        var_name : str
            Name of variable in file.
        vars_as_series : bool
            If True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects.

        Returns
        -------
        StationData
            Dict-like object containing the results.

        """
        if var_name not in self.PROVIDES_VARIABLES:
            raise ValueError(f"Invalid input variable {var_name}")

        # there's only one variable in the file
        aerocom_var_name = var_name

        # Iterate over the lines of the file
        logger.info(f"Reading file {filename}")
        file_delimiter = self.FILE_COL_DELIM
        # this lists the data to keep from the original read string
        # this becomes a time series
        file_indexes_to_keep = [11, 13, 14, 15, 16]
        # used for line length control...
        max_file_index_to_keep = max(file_indexes_to_keep)
        # this is some header information
        header_indexes_to_keep = [
            0,
            3,
            8,
            9,
            10,
            12,
        ]
        # These are the indexes with a time and are stored as np.datetime64
        time_indexes = [13, 14]

        # read the file
        # enable alternative reading of .gz files here to save space on the file system
        suffix = pathlib.Path(filename).suffix
        if suffix == ".gz":
            f_out = tempfile.NamedTemporaryFile(delete=False)
            with gzip.open(filename, "r") as f_in:
                shutil.copyfileobj(f_in, f_out)
            read_filename = f_out.name
            f_out.close()
        else:
            read_filename = filename

        # input files can be either UTF-8 or UTF-16 encoded
        # try both
        # files are max 3MB in size, so no big deal terms of RAM usage
        try:
            with open(read_filename) as f:
                lines = f.readlines()
        except UnicodeDecodeError:
            with open(read_filename, encoding="UTF-16") as f:
                lines = f.readlines()
        except Exception:
            if suffix == ".gz":
                os.remove(f_out.name)
            raise EEAv2FileError(f"Found corrupt file {filename}. consider deleting it")

        # remove the temp file in case the input file was a gz file
        if suffix == ".gz":
            os.remove(f_out.name)

        header = lines[0].lower().rstrip().split(file_delimiter)
        # create output dict
        if len(header) < max_file_index_to_keep:
            raise EEAv2FileError(f"Found corrupt file {filename}. consider deleting it")

        data_dict = {}
        for idx in header_indexes_to_keep:
            data_dict[header[idx]] = ""

        for idx in file_indexes_to_keep:
            if idx in time_indexes:
                data_dict[header[idx]] = np.zeros(self.MAX_LINES_TO_READ, dtype="datetime64[s]")
            else:
                data_dict[header[idx]] = np.empty(self.MAX_LINES_TO_READ, dtype=np.float64)

        # read the data...
        # DE,http://gdi.uba.de/arcgis/rest/services/inspire/DE.UBA.AQD,NET.DE_BB,STA.DE_DEBB054,DEBB054,SPO.DE_DEBB054_PM2_dataGroup1,SPP.DE_DEBB054_PM2_automatic_light-scat_Duration-30minute,SAM.DE_DEBB054_2,PM2.5,http://dd.eionet.europa.eu/vocabulary/aq/pollutant/6001,hour,3.2000000000,µg/m3,2020-01-04 00:00:00 +01:00,2020-01-04 01:00:00 +01:00,1,2
        lineidx = 0
        for line in lines[1:]:
            rows = line.rstrip().split(file_delimiter)
            # Unfortunately there's a lot of corrupt files
            # skip data line if the # rows is not sufficient
            if len(rows) < max_file_index_to_keep:
                continue
            if lineidx == 0:
                for idx in header_indexes_to_keep:
                    if header[idx] != self.VAR_CODE_NAME:
                        data_dict[header[idx]] = rows[idx]
                    else:
                        # extract the EEA var code from the URL noted in the data file
                        data_dict[header[idx]] = rows[idx].split("/")[-1]

            for idx in file_indexes_to_keep:
                # if the data is a time
                if idx in time_indexes:
                    # make the time string ISO compliant so that numpy can directly read it
                    # this is not very time string forgiving but fast
                    data_dict[header[idx]][lineidx] = np.datetime64(
                        rows[idx][0:10] + "T" + rows[idx][11:19]
                    )
                    # due to the deprecation of the timezone interpretation after numpy 0.11
                    # we have to subtract the offset manually to get to UTC.
                    # np.timedelta64 does not accept a float as parameter, only an integer.
                    # Although there are time zones with a 30 minutes offset, these don't
                    # exist in Europe, so just consider integer hours here for speed
                    tz_offset = np.timedelta64(np.int64(rows[idx][20:23]), "h")
                    data_dict[header[idx]][lineidx] = data_dict[header[idx]][lineidx] - tz_offset
                else:
                    # data is not a time
                    # sometimes there's no value in the file. Set that to nan
                    try:
                        data_dict[header[idx]][lineidx] = np.float64(rows[idx])
                    except (ValueError, IndexError):
                        data_dict[header[idx]][lineidx] = np.nan

            lineidx += 1
        # if the first line in the file was empty
        if data_dict["unitofmeasurement"] == "":
            if rows[12] == "":
                raise EEAv2FileError(
                    f"Unit of Measurement could not be inferred from EEA file {filename}"
                )
            else:
                # with loss of generality get the unitofmeasurement from the last row column 12 (which should be a kept header)
                data_dict["unitofmeasurement"] = rows[12]

        unit_in_file = data_dict["unitofmeasurement"]
        # adjust the unit and apply conversion factor in case we read a variable noted in self.AUX_REQUIRES
        if var_name in self.AUX_REQUIRES:
            unit_in_file = self.CONV_UNIT[var_name]
            data_dict[self.VAR_NAMES_FILE[var_name]] = (
                data_dict[self.VAR_NAMES_FILE[var_name]] * self.CONV_FACTOR[var_name]
            )
        try:
            unit = self.VAR_UNITS_FILE[unit_in_file]
        except KeyError:
            # this will raise an Exception if pyaerocom.units cannot handle. In
            # which case the unit should be added in VAR_UNITS_FILE
            unit = str(Unit(unit_in_file))

        # Empty data object (a dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id
        data_out.dataset_name = self.DATASET_NAME
        data_out.station_id = data_dict["airqualitystation"]
        data_out.station_name = data_dict["airqualitystation"]
        data_out.filename = filename
        data_out.instrument_name = self.INSTRUMENT_NAME
        data_out.country_code = data_dict["countrycode"]
        freq = data_dict["averagingtime"]
        try:
            tstype = self.TS_TYPES_FILE[freq]
        except KeyError:
            raise TemporalResolutionError(
                f"Found invalid ts_type {freq}. Please register in class header attr TS_TYPES_FILE"
            )

        data_out.ts_type = tstype
        # ToDo: check "variables" entry, it should not be needed anymore in UngriddedData
        data_out["variables"] = [aerocom_var_name]
        data_out["var_info"][aerocom_var_name] = {}
        data_out["var_info"][aerocom_var_name]["units"] = unit
        # TsType is
        # data_out['var_info'][aerocom_var_name]['ts_type'] = self.TS_TYPE

        # Sometimes the times in the data files are not ordered in time which causes problems when doing
        # time interpolations later on. Make sure that the data is ordered in time
        diff_unsorted = np.diff(data_dict[self.START_TIME_NAME])
        sort_flag = False
        # use a vectorised time_diff instead of scalar one as before
        time_diff = (
            data_dict[self.END_TIME_NAME][:lineidx] - data_dict[self.START_TIME_NAME][:lineidx]
        ) / 2.0

        # np.min needs an array and fails with ValueError when a scalar is supplied
        # this is the case for a single line file
        try:
            min_diff = np.min(diff_unsorted)
        except ValueError:
            min_diff = 0

        if min_diff < 0:
            # data needs to be sorted
            ordered_idx = np.argsort(data_dict[self.START_TIME_NAME][:lineidx])
            data_out["dtime"] = (
                data_dict[self.START_TIME_NAME][ordered_idx] + time_diff[ordered_idx]
            )
            sort_flag = True
        else:
            data_out["dtime"] = data_dict[self.START_TIME_NAME][:lineidx] + time_diff

        for key, value in data_dict.items():
            # adjust the variable name to aerocom standard
            if key != self.VAR_NAMES_FILE[aerocom_var_name]:
                data_out[key] = value[:lineidx]
            else:
                if sort_flag:
                    data_out[aerocom_var_name] = value[:lineidx][ordered_idx]
                else:
                    data_out[aerocom_var_name] = value[:lineidx]

        # convert data vectors to pandas.Series (if attribute
        # vars_as_series=True)
        if vars_as_series:
            data_out[aerocom_var_name] = pd.Series(
                data_out[aerocom_var_name], index=data_out["dtime"]
            )

        return data_out

    def _read_metadata_file(self, filename=None):
        """Read EEA metadata file

        Parameters
        ----------
        filename : str
            Absolute path to filename to read.

        Returns
        -------
        metadata
            Dict-like object containing the results with the keys being a combination of the
            station name and the variable and the values being all fields of the metadata file

        """

        if filename is None:
            filename = os.path.join(self.data_dir, self.DEFAULT_METADATA_FILE)
            # test also for a gzipped file...
        if not os.path.isfile(filename):
            filename = filename + ".gz"
        logger.warning(f"Reading file {filename}")

        struct_data = {}
        suffix = pathlib.Path(filename).suffix
        if suffix == ".gz":
            f_out = tempfile.NamedTemporaryFile(delete=False)
            with gzip.open(filename, "r") as f_in:
                shutil.copyfileobj(f_in, f_out)
            filename = f_out.name
            f_in.close()

        with open(filename) as f:
            # read header...
            # Countrycode Timezone Namespace   AirQualityNetwork AirQualityStation AirQualityStationEoICode   AirQualityStationNatCode   SamplingPoint  SamplingProcess Sample   AirPollutantCode  ObservationDateBegin ObservationDateEnd   Projection  Longitude   Latitude Altitude MeasurementType   AirQualityStationType   AirQualityStationArea   EquivalenceDemonstrated MeasurementEquipment InletHeight BuildingDistance  KerbDistance
            header = f.readline().lower().rstrip().split()
            min_row_no = len(header)
            # create output dict
            data_dict = {}
            for key in header:
                data_dict[key] = []
            lineidx = 0
            bad_line_no = 0
            bad_line_arr = []
            for line in f:
                rows = line.rstrip().split("\t")
                # skip too short lines
                if len(rows) < min_row_no:
                    # print(line)
                    bad_line_no += 1
                    bad_line_arr.append(line)
                    continue
                temp_dict = {}
                for idx, key in enumerate(header):
                    if header[idx] != self.VAR_CODE_NAME:
                        # data_dict[header[idx]] = rows[idx]
                        temp_dict[header[idx]] = rows[idx]
                    else:
                        # extract the EEA var code from the URL noted in the data file
                        # data_dict[header[idx]] = rows[idx].split('/')[-1]
                        temp_dict[header[idx]] = rows[idx].split("/")[-1]

                meta_key = f"{temp_dict['airqualitystation']}__{temp_dict['airpollutantcode']}"
                if meta_key not in struct_data:
                    struct_data[meta_key] = temp_dict.copy()
                else:
                    pass
                lineidx += 1

        logger.info(f"Reading file {filename} done")
        # remove the temp file in case the input file was a gz file
        if suffix == ".gz":
            f_out.close()
            os.remove(f_out.name)

        return struct_data

    def get_file_list(self, pattern=None):
        """Search all files to be read

        Uses :attr:`_FILEMASK` (+ optional input search pattern, e.g.
        station_name) to find valid files for query.

        Parameters
        ----------
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing retrieved file locations

        Raises
        ------
        IOError
            if no files can be found
        """

        if pattern is None:
            logger.warning("using default pattern *.* for file search")
            pattern = "*.*"

        logger.info("Fetching data files. This might take a while...")
        fp = os.path.join(self.data_dir, pattern)
        return sorted(glob.glob(fp, recursive=True))

    def get_station_coords(self, meta_key):
        """
        get a station's coordinates

        Parameters
        ----------
        meta_key : str
            string with the internal station key
        """
        ret_data = {}
        ret_data["latitude"] = float(self._metadata[meta_key][self.LATITUDENAME])
        ret_data["longitude"] = float(self._metadata[meta_key][self.LONGITUDENAME])
        ret_data["altitude"] = float(self._metadata[meta_key][self.ALTITUDENAME])
        return ret_data

    def read(
        self, vars_to_retrieve=None, files=None, first_file=None, last_file=None, metadatafile=None
    ):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional
            List containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded.
        files : :obj:`list`, optional
            List of files to be read. If None, then the file list used is the
            returned from :func:`get_file_list`.
        first_file : :obj:`int`, optional
            Index of the first file in :obj:'file' to be read. If None, the
            very first file in the list is used.
        last_file : :obj:`int`, optional
            Index of the last file in :obj:'file' to be read. If None, the very
            last file in the list is used.
        metadatafile : :obj:'str', optional
            full qualified path to metadata file. If None, the default metadata
            file will be used

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if len(vars_to_retrieve) > 1:
            raise NotImplementedError("So far, only one variable can be read at a time...")
        var_name = vars_to_retrieve[0]
        logger.info("Reading EEA data")
        if files is None:
            logger.info("Retrieving file list")
            files = self.get_file_list(self.FILE_MASKS[var_name])

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        if metadatafile is None:
            metadatafile = os.path.join(self.data_dir, self.DEFAULT_METADATA_FILE)

        files = files[first_file:last_file]

        data_obj = UngriddedData()
        meta_key = 0.0
        idx = 0

        # Assign metadata object and index
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        logger.info("Reading metadata file")
        # non compliant, but efficiently indexed metadata
        self._metadata = self._read_metadata_file(metadatafile)

        # returns a dict with country codes as keys and the country names as value
        _country_dict = get_country_name_from_iso()
        logger.info("Reading files...")

        for i in tqdm(range(len(files)), disable=None):
            _file = files[i]
            try:
                station_data = self.read_file(_file, var_name=var_name)
            except EEAv2FileError:
                logger.warning(f"file {_file} is corrupt! consider deleting it")
                continue
            except TemporalResolutionError as e:
                logger.warning(f"{_file} has TemporalResolutionError")
                logger.warning(f"{repr(e)}. Skipping file...")
                continue

            # readfile might fail outside of the error captured by the try statement above
            if station_data is None:
                logger.warning(f"file {_file} did not provide data. skipping...!")
                continue

            # to find the metadata quickly, we use a string internally
            _meta_key = f"{station_data['station_id']}__{station_data['airpollutantcode']}"

            # Fill the metadata dict.
            # The location in the data set is time step dependant
            if _meta_key not in self._metadata:
                logger.warning(
                    f"metadata for station {_meta_key} not found! skipping that station!"
                )
                continue
            metadata[meta_key] = {}
            meta_idx[meta_key] = {}
            metadata[meta_key].update(station_data.get_meta())
            metadata[meta_key].update(self.get_station_coords(_meta_key))
            metadata[meta_key]["variables"] = list(
                station_data.var_info.keys()
            )  # vars_to_retrieve
            metadata[meta_key]["station_classification"] = self._metadata[_meta_key][
                "airqualitystationtype"
            ]
            metadata[meta_key]["area_classification"] = self._metadata[_meta_key][
                "airqualitystationarea"
            ]
            try:
                metadata[meta_key]["country"] = _country_dict[metadata[meta_key]["country_code"]]
            except KeyError:
                pass
            metadata[meta_key]["var_info"] = station_data["var_info"]
            metadata[meta_key]["website"] = self.WEBSITE
            metadata[meta_key]["data_product"] = self.DATA_PRODUCT
            metadata[meta_key]["station_name"] = self._metadata[_meta_key][
                "airqualitystationeoicode"
            ]

            # List with indices of this station for each variable
            num_times = len(station_data["dtime"])

            # Check whether the size of the data object needs to be extended
            if (idx + num_times) >= data_obj._ROWNO:
                # if totnum < data_obj._CHUNKSIZE, then the latter is used
                data_obj.add_chunk(num_times)

            for var_idx, var in enumerate(list(station_data.var_info)):
                # set invalid data to np.nan according to
                # https://dd.eionet.europa.eu/vocabulary/aq/observationvalidity/view
                # data flagged as below the detection limit (values 2 and 3)
                # will remain in the data
                station_data[var][station_data["validity"] < 1] = np.nan

                values = station_data[var]
                start = idx + var_idx * num_times
                stop = start + num_times

                data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key
                data_obj._data[start:stop, data_obj._TIMEINDEX] = station_data["dtime"]
                data_obj._data[start:stop, data_obj._DATAINDEX] = values
                data_obj._data[start:stop, data_obj._VARINDEX] = var_idx
                meta_idx[meta_key][var] = np.arange(start, stop)

                if var not in data_obj.var_idx:
                    data_obj.var_idx[var] = var_idx

            idx += num_times
            meta_key = meta_key + 1.0

        # Shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]
        # data_obj.data_revision[self.DATASET_NAME] = self.data_revision
        self._metadata = None
        self.files = files

        return data_obj



from .molecular_mass import MolecularMass

# Definitions of various custom constants and conversion factors.

M_H = MolecularMass("H")
M_O = MolecularMass("O")
M_N = MolecularMass("N")
M_S = MolecularMass("S")

M_SO2 = MolecularMass("SO2")
M_SO4 = MolecularMass("SO4")
M_NO2 = MolecularMass("NO2")
M_NO3 = MolecularMass("NO3")

M_NH3 = MolecularMass("NH3")
M_NH4 = MolecularMass("NH4")

HA_TO_SQM = 10_000  # m^2 ha^1

SECONDS_IN_DAY = 24 * 60 * 60



from .reporting_base import get_CFG as get_EMEP_CFG


def get_CFG(year: int, model_dir: str, *, file_pattern: str = r"^RERUN2022_{freq}_.+\.nc$"):
    """Basically the EMEP base config with minor changes
    to work with EMEP4NO input. Please also refer to the
    EMEP config documentation as that is where the bulk
    of the configuration takes place.

    :param year: Year of data
    :param model_dir: Directory where EMEP4NO files are located.
    :param file_pattern: Optional regular expression against which the base name of files will
        be matched. This can be used to override the default `Base_{freq}.nc` file matching in
        the EMEP reader.

        Note that for convenience the string literal '{freq}' can be included as part of the
        pattern and will be expanded to (hour|day|month|fullrun). This is recommended, as
        the presence of these strings are used to derive ts_type, which is currently necessary
        for reading.

    :returns: A dict of model configuration which can be passed to EvalSetup.

    Example
    -------------

    The following snippet shows how this config can be used.

    >>> from pyaerocom.aeroval.config.emep.emep4no_base_config import get_CFG # doctest: +SKIP
    >>> import pathlib
    >>>
    >>> if __name__ == "__main__":
    ...     import matplotlib.pyplot as plt
    ...     import pyaerocom as pya
    ...     from pyaerocom import const
    ...     from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
    ...
    ...     # Customize cache dir to avoid disk quota issues.
    ...     # cdir = pathlib.Path("./cache")
    ...     # cdir.mkdir(exist_ok=True)
    ...     # const.CACHEDIR = str(cdir)
    ...
    ...     cfg = get_CFG(2022, "/lustre/storeB/project/fou/kl/emep/ModelRuns/EMEP4NO/EMEP4NO_rerun_2022/")
    ...
    ...     # Change any experiment details.
    ...     cfg.update(
    ...         {
    ...             #"proj_id": "<project name>",
    ...             #"exp_id": "<experiment name>",
    ...             #"json_basedir": "/lustre/storeB/users/thlun8736/python/aeroval/data",
    ...             #"coldata_basedir": "/lustre/storeB/users/thlun8736/python/aeroval/coldata",
    ...         }
    ...     )
    ...
    ...     # Run the experiment.
    ...     stp = EvalSetup(**cfg)
    ...     ana = ExperimentProcessor(stp)
    ...     res = ana.run()
    """
    cfg = get_EMEP_CFG(None, year, model_dir)

    # Ensure files are matched properly by overriding 'Base_{freq}.nc'
    # matching in reader.
    cfg["model_cfg"]["EMEP"]["model_kwargs"] = {"file_pattern": file_pattern}

    cfg.update(
        {
            "exp_id": "emep4no",
            "exp_name": "Evaluation of EMEP4NO runs",
            "exp_descr": "Evaluation of EMEP4NO runs",
            "add_model_maps": False,
            "raise_exceptions": False,
        }
    )
    return cfg



from pyaerocom.io.mscw_ctm.additional_variables import (
    calc_concNhno3,
    calc_concNnh3,
    calc_concNnh4,
    calc_concNno3pm10,
    calc_concNno3pm25,
    calc_concNtnh,
    calc_conNtno3,
    calc_ratpm10pm25,
    calc_ratpm25pm10,
    update_EC_units,
)
from tests.fixtures.mscw_ctm import create_fake_MSCWCtm_data
from pyaerocom.units.constants import M_N, M_O, M_H


def test_calc_concNhno3():
    conchno3 = create_fake_MSCWCtm_data()

    concNhno3_from_func = calc_concNhno3(conchno3)

    concNhno3 = conchno3 * (M_N / (M_H + M_N + M_O * 3))
    concNhno3.attrs["units"] = "ug N m-3"
    assert (concNhno3 == concNhno3_from_func).all()


def test_calc_concNno3pm10():
    concno3c = create_fake_MSCWCtm_data()
    concno3f = create_fake_MSCWCtm_data()

    concNno3pm10_from_func = calc_concNno3pm10(concno3f, concno3c)

    fac = M_N / (M_N + 3 * M_O)
    concno3pm10 = concno3f + concno3c
    concNno3pm10 = concno3pm10 * fac
    concNno3pm10.attrs["var_name"] = "concNno3pm10"
    concNno3pm10.attrs["units"] = "ug N m-3"
    assert (concNno3pm10 == concNno3pm10_from_func).all()


def test_calc_concNno3pm25():
    concno3c = create_fake_MSCWCtm_data()
    concno3f = create_fake_MSCWCtm_data()

    concNno3pm10_from_func = calc_concNno3pm25(concno3f, concno3c)

    fac = M_N / (M_N + 3 * M_O)
    concno3pm10 = concno3f + 0.134 * concno3c
    concNno3pm10 = concno3pm10 * fac
    concNno3pm10.attrs["var_name"] = "concNno3pm10"
    concNno3pm10.attrs["units"] = "ug N m-3"
    assert (concNno3pm10 == concNno3pm10_from_func).all()


def test_calc_conNtno3():
    conchno3 = create_fake_MSCWCtm_data()
    concno3f = create_fake_MSCWCtm_data()
    concno3c = create_fake_MSCWCtm_data()

    concNtno3_from_func = calc_conNtno3(conchno3, concno3f, concno3c)

    concNhno3 = calc_concNhno3(conchno3)
    concNno3pm10 = calc_concNno3pm10(concno3f, concno3c)

    concNtno3 = concNhno3 + concNno3pm10
    concNtno3.attrs["units"] = "ug N m-3"
    assert (concNtno3 == concNtno3_from_func).all()


def test_calc_concNtnh():
    concnh3 = create_fake_MSCWCtm_data()
    concnh4 = create_fake_MSCWCtm_data()

    concNtnh_from_func = calc_concNtnh(concnh3, concnh4)

    concNnh3 = calc_concNnh3(concnh3)
    concNnh4 = calc_concNnh4(concnh4)

    concNtnh = concNnh3 + concNnh4
    concNtnh.attrs["units"] = "ug N m-3"
    assert (concNtnh == concNtnh_from_func).all()


def test_calc_concpm10pm25():
    concpm10 = create_fake_MSCWCtm_data()
    concpm25 = create_fake_MSCWCtm_data()

    ratpm10pm25_from_func = calc_ratpm10pm25(concpm10, concpm25)
    assert ratpm10pm25_from_func.attrs["units"] == "1"


def test_calc_concpm25pm10():
    concpm10 = create_fake_MSCWCtm_data()
    concpm25 = create_fake_MSCWCtm_data()

    ratpm25pm10_from_func = calc_ratpm25pm10(concpm25, concpm10)
    assert ratpm25pm10_from_func.attrs["units"] == "1"


def test_calc_concNnh3():
    concnh3 = create_fake_MSCWCtm_data()

    concNnh3_from_func = calc_concNnh3(concnh3)

    concNnh3 = concnh3 * (M_N / (M_H * 3 + M_N))
    concNnh3.attrs["units"] = "ug N m-3"
    assert (concNnh3 == concNnh3_from_func).all()


def test_calc_concNnh4():
    concnh4 = create_fake_MSCWCtm_data()

    concNnh4_from_func = calc_concNnh4(concnh4)

    concNnh4 = concnh4 * (M_N / (M_H * 4 + M_N))
    concNnh4.attrs["units"] = "ug N m-3"
    assert (concNnh4 == concNnh4_from_func).all()


def test_update_EC_units():
    concecpm25 = create_fake_MSCWCtm_data()

    concCecpm25_from_func = update_EC_units(concecpm25)

    concCecpm25 = concecpm25
    concCecpm25.attrs["units"] = "ug C m-3"

    assert (concCecpm25 == concCecpm25_from_func).all()
    assert concCecpm25.units == concCecpm25_from_func.units



from pyaerocom.aeroval.config.pmratios.base_config import get_CFG


def test_ratpmconfig():
    """short test if the example configuration for pm ratios is still in the code"""

    reportyear = year = 2019
    CFG = get_CFG(
        reportyear=reportyear,
        year=year,
        model_dir="/lustre/storeB/project/fou/kl/CAMEO/u8_cams0201/",
    )
    assert not CFG["raise_exceptions"]



from __future__ import annotations

import logging
import os
from datetime import datetime
from pathlib import Path

import numpy as np
from tqdm import tqdm

from pyaerocom import const
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.io.icpforests.metadata import MetadataReader, Station, SurveyYear
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData

logger = logging.getLogger(__name__)


class ReadICPForest(ReadUngriddedBase):
    #: version log of this class (for caching)
    __version__ = "0.5_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.ICPFORESTS_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.ICPFORESTS_NAME]

    TS_TYPE = "undefined"

    _FILEMASK = "dp_dem.csv"

    #: Temporal resolution codes that (so far) can be understood by pyaerocom
    TS_TYPE_CODES = {
        "1mn": "minutely",
        "1h": "hourly",
        "1d": "daily",
        "1w": "weekly",
        "1mo": "monthly",
        "mn": "minutely",
        "h": "hourly",
        "d": "daily",
        "w": "weekly",
        "mo": "monthly",
    }

    VAR_POSITION = {
        "wetoxs": 20,
        "wetoxn": 19,
        "wetrdn": 17,
        "dryoxs": 20,
        "dryoxn": 19,
        "dryrdn": 17,
        "depoxs": 20,
        "depoxn": 19,
        "deprdn": 17,
        "depna": 16,
        "depoxsf": 20,
        "depoxnf": 19,
        "deprdnf": 17,
        "depnaf": 16,
        "wetso4": 20,
        "wetno3": 19,
        "wetnh4": 17,
        "wetno2": 41,
        "wetna": 16,
        "wetcl": 18,
    }

    UNITS = {
        "wetoxs": "mg S m-2 d-1",
        "wetoxn": "mg N m-2 d-1",
        "wetrdn": "mg N m-2 d-1",
        "dryoxs": "mg S m-2 d-1",
        "dryoxn": "mg N m-2 d-1",
        "dryrdn": "mg N m-2 d-1",
        "depoxs": "mg S m-2 d-1",
        "depoxn": "mg N m-2 d-1",
        "deprdn": "mg N m-2 d-1",
        "depoxsf": "mg S m-2 d-1",
        "depoxnf": "mg N m-2 d-1",
        "deprdnf": "mg N m-2 d-1",
        "wetso4": "mg S m-2 d-1",
        "wetno3": "mg N m-2 d-1",
        "wetnh4": "mg N m-2 d-1",
        "wetno2": "mg N m-2 d-1",
        "wetna": "mg m-2 d-1",
        "depna": "mg m-2 d-1",
        "depnaf": "mg m-2 d-1",
        # unverified
        "wetcl": "mg m-2 d-1",
    }

    SEASALT_CORRECTION = {
        "wetoxs": 0.3338,
        "wetso4": 0.3338,
        "depoxs": 0.3338,
        "depoxsf": 0.3338,
    }
    SEASALT_FACTORS = {
        "wetna": 0.120,
        "wetcl": 0.103,
    }

    DEP_TYPES_TO_USE = ["Throughfall", "Bulk", "Wet-only"]

    QUALITY_LIMIT = 0.5

    MIN_YEAR = 1984
    MAX_YEAR = 2019

    def __init__(self, data_id=None, data_dir=None):
        super().__init__(data_id, data_dir)

        self.metadata = None

        self._file_dir = None
        self.files = None

        if data_dir is not None:
            self.metadata = MetadataReader(data_dir)

    @property
    def file_dir(self):
        """Directory containing EBAS NASA Ames files"""
        if self._file_dir is not None:
            return self._file_dir
        return os.path.join(self.data_dir)

    @file_dir.setter
    def file_dir(self, val):
        if not isinstance(val, str) or not os.path.exists(val):
            raise FileNotFoundError("Input directory does not exist")

        self._file_dir = val

    def read(self, vars_to_retrieve=None, files=[], first_file=None, last_file=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used

        Returns
        -------
        UngriddedData
            instance of ungridded data object containing data from all files.
        """
        self.files = Path(self.data_dir).joinpath(self._FILEMASK)
        data = self.read_file(
            str(self.files),
            vars_to_retrieve,
        )

        return data

    @property
    def PROVIDES_VARIABLES(self):
        """List of variables that are provided by this dataset

        Note
        ----
        May be implemented as global constant in header
        """
        return list(self.VAR_POSITION.keys())

    @property
    def DEFAULT_VARS(self):
        """List containing default variables to read"""
        return list(self.VAR_POSITION.keys())

    def read_file(self, filename, vars_to_retrieve=None, last_line=None):
        """Read single file

        Parameters
        ----------
        filename : str
            string specifying filename
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        last_line : int
            last line number to read
            Used to speed up testing

        Returns
        -------
        :obj:`dict` or :obj:`StationData`, or other...
            imported data in a suitable format that can be handled by
            :func:`read` which is supposed to append the loaded results from
            this method (which reads one datafile) to an instance of
            :class:`UngriddedData` for all files.
        """
        stations: dict[str, dict[str, Station]] = {}
        if self.metadata is None:
            if self.data_dir is None:
                raise ValueError("Data Dir is not read yet")

            self.metadata = MetadataReader(self.data_dir)

        if vars_to_retrieve is None:
            vars_to_retrieve = self.PROVIDES_VARIABLES

        with open(filename) as f:
            f.readline()

            for line_nr, line in tqdm(enumerate(f)):
                words = line.split(";")
                year = int(words[0])
                country_code = int(words[1])
                partner_code = int(words[2])
                plot_code = int(words[3])
                sampler_code = int(words[9])
                q_flag = int(words[44]) if words[44] != "" else 0

                # 8 is the code for "do not use"
                if (
                    self.metadata.deposition_type[sampler_code] not in self.DEP_TYPES_TO_USE
                    or sampler_code == 8
                ):
                    continue

                period = int(words[6])
                start = words[4]
                stop = words[5]

                quantity = words[47]
                if quantity == "":  # or quantity == "0":
                    continue
                else:
                    quantity = float(quantity)

                try:
                    self.metadata.plots.plots[country_code]
                    self.metadata.plots.plots[country_code][plot_code]
                    self.metadata.plots.plots[country_code][plot_code][sampler_code]
                except KeyError:
                    logger.warning(
                        f"Some metadata is missing for {country_code=}, {plot_code=}, {sampler_code=}. Skipping"
                    )
                    continue

                try:
                    self.metadata.plots.plots[country_code][plot_code][sampler_code].survey_years[
                        year
                    ]
                except KeyError:
                    logger.warning(
                        f"Year {year} can't be found for {country_code=}, {plot_code=}, {sampler_code=}. Only years found are {self.metadata.plots.plots[country_code][plot_code][sampler_code].survey_years.keys()}"
                    )
                    continue
                days, dtime, ts_type = self._get_days_date_ts_type(
                    year, country_code, plot_code, sampler_code, period, start, stop
                )

                if days is None or dtime is None or ts_type is None:
                    continue
                station_name = Station.get_station_name(country_code, plot_code, sampler_code)

                if station_name not in stations:
                    stations[station_name] = {}
                if ts_type not in stations[station_name]:
                    lat, lon, alt = self.metadata.plots.get_position(
                        year, country_code, plot_code, sampler_code
                    )
                    stations[station_name][ts_type] = Station(
                        country_code, plot_code, sampler_code, lat, lon, alt, partner_code, ts_type
                    )

                for species in vars_to_retrieve:
                    conc = self._get_species_conc(words[self.VAR_POSITION[species]], species)

                    # Sea-salt correction
                    # The factor self.SEASALT_CORRECTION[species] is the factor use to go from mg/L to mg S/L (for sulpher)
                    if species in self.SEASALT_CORRECTION:
                        na_factor = (
                            self._get_species_conc(words[self.VAR_POSITION["wetna"]], "wetna")
                            * self.SEASALT_FACTORS["wetna"]
                            * self.SEASALT_CORRECTION[species]
                        )
                        cl_factor = (
                            self._get_species_conc(words[self.VAR_POSITION["wetcl"]], "wetcl")
                            * self.SEASALT_FACTORS["wetcl"]
                            * self.SEASALT_CORRECTION[species]
                        )
                        seasalt_correction = min(na_factor, cl_factor)

                        if seasalt_correction > conc and conc > 0:
                            logger.warning(
                                f"Seasalt correction {seasalt_correction} is larger than concentration {conc} for {species}"
                            )
                        if conc > 0:
                            conc -= seasalt_correction

                    # Unit correction
                    conc *= quantity / days

                    stations[station_name][ts_type].add_measurement(
                        species, dtime, conc, self.UNITS[species], q_flag
                    )

        station_datas = []
        for station_name in stations:
            for ts_type in stations[station_name]:
                station = stations[station_name][ts_type]
                station_data = StationData()
                station_data.var_info = BrowseDict(**station.var_info)
                for species in station.data.keys():
                    station_data[species] = self._clean_data_with_flags(
                        station.data[species],
                        station.dtime[species],
                        station.flags[species],
                        species,
                    )
                    station_data.dtime = station.dtime[species]

                station_data.country = station.country

                # Needs to convert coordinates to correct type!
                station_data.station_coords = {
                    "latitude": station.lat,
                    "longitude": station.lon,
                    "altitude": station.alt,
                }

                station_data.latitude = station.lat
                station_data.longitude = station.lon
                station_data.altitude = station.alt

                station_data.filename = filename
                station_data.ts_type = station.ts_type
                station_data.ts_type_src = station.ts_type
                station_data.station_name = station.station_name
                station_data._append_meta_item("sampler_type", station.sampler_type)

                station_data.data_id = self.data_id

                station_datas.append(station_data)
        return UngriddedData.from_station_data(station_datas, add_meta_keys="sampler_type")

    def _get_species_conc(self, conc_str: str, species: str) -> float:
        conc = float(conc_str) if conc_str != "" else np.nan
        if conc == -1:
            logger.warning(f"Value for {species} found to be {conc}")
            conc = np.nan
        return conc

    def _get_days_date_ts_type(
        self,
        year: int,
        country_code: int,
        plot_code: int,
        sampler_code: int,
        period: int,
        start: str | datetime,
        stop: str | datetime,
    ) -> tuple[float | None, datetime | None, str | None]:
        if start != "" and stop != "":
            if isinstance(start, str):
                start = datetime.strptime(start, "%Y-%m-%d")
            if isinstance(stop, str):
                stop = datetime.strptime(stop, "%Y-%m-%d")

            if (stop - start).days <= 0:
                return None, None, None

            return (stop - start).days, start, self._get_tstype(start, stop)

        if self.metadata is None:
            raise ValueError("Metadata is not read yet")

        try:
            days = self.metadata.plots.get_days(year, country_code, plot_code, sampler_code)
        except ValueError as e:
            logger.warning(repr(e))
            return None, None, None

        if days == 0:
            return None, None, None

        try:
            dtime = self.metadata.plots.get_date(
                year, country_code, plot_code, sampler_code, period
            )
        except ValueError:
            return None, None, None

        ts_type = self.metadata.plots.get_ts_type(year, country_code, plot_code, sampler_code)

        return days, dtime, ts_type

    def _get_tstype(self, start: datetime, stop: datetime) -> str:
        days = (stop - start).days

        return SurveyYear.get_tstype(days)

    def _clean_data_with_flags(
        self,
        data: list[float],
        time: list[datetime],
        flags: list[int],
        species: str,
    ) -> list[float]:
        data_array = np.array(data)
        flags_array = np.array(flags)
        years = np.array([i.year for i in time])

        for year in range(self.MIN_YEAR, self.MAX_YEAR):
            yr_flags = flags_array[np.where(years == year)]
            # yr_flags can have 0 length
            try:
                quality = np.sum(np.where(yr_flags == 0)) / len(yr_flags)
                if quality < self.QUALITY_LIMIT:
                    logger.warning(
                        f"Quality of {quality} found for {species} in year {year}. Setting data this year to NaN"
                    )
                    data_array[np.where(years == year)] = np.nan

            except RuntimeWarning:
                # yr_flags has 0 length
                logger.warning(f"No data for species {species} in year {year}.")

        return list(data_array)



from pathlib import Path

import pytest

from pyaerocom.io.cams2_83.models import ModelData, ModelName, RunType


@pytest.mark.parametrize("model", ModelName)
def test_ModelName(model: ModelName):
    assert model == str(model) == model.value


@pytest.mark.parametrize("run", RunType)
def test_RunType(run: RunType):
    assert run == str(run) == run.value


@pytest.fixture
def model_data(tmp_path: Path, model_name: ModelName, run_type: RunType) -> ModelData:
    """ModelData on a temporary path"""
    return ModelData(model_name, run_type, root=tmp_path)


@pytest.mark.parametrize("model_name", ModelName)
@pytest.mark.parametrize("run_type", RunType)
def test_ModelData_frompath(model_data: ModelData):
    path = model_data.path
    assert ModelData.frompath(str(path)) == ModelData.frompath(path) == model_data



#: Default variable ranges for web display
import copy
import json
import logging
from configparser import ConfigParser
from enum import Enum
import os
from typing import NamedTuple

from pydantic import BaseModel

from pyaerocom.data import resources

logger = logging.getLogger(__name__)


# basemodel-implementation for verification
class _ScaleAndColmap(NamedTuple):
    scale: list[float]
    colmap: str


class _VarWebScaleAndColormap(BaseModel):
    scale_colmaps: dict[str, _ScaleAndColmap]


# dict-implementation for json-serialization, namedtuple not stable with json/simplejson
class ScaleAndColmap(dict[str, str | list[float]]):
    """simple dictionary container with only two keys, scale and colmap"""

    pass


class VarWebScaleAndColormap(dict[str, ScaleAndColmap]):
    def __init__(self, config_file: str = "", **kwargs):
        """This class contains scale and colmap information and is implemented as dict to allow
        json serialization. It reads it initial data from data/var_scale_colmap.ini. kwargs will be send to update.

        :param config_file: filename to additional or updated information, defaults to None
        """
        super().__init__()
        with resources.path("pyaerocom.aeroval.data", "var_scale_colmap.ini") as file:
            self.update_from_ini(file)
        if config_file != "":
            logger.info(f"Reading additional web-scales from '{config_file}'")
            if not os.path.exists(config_file):
                raise FileNotFoundError(
                    f"VarWebScaleAndColormap initialized with config_file: '{config_file}' which does not exist"
                )
            self.update_from_ini(config_file)
        self.update(**kwargs)

    def update(self, **kwargs):
        """update/add scale and colormaps by kwargs, e.g.
        update(concso2={"scale"=[0.2,1], "colormap"="bluewhite"})
        """
        wvsc = _VarWebScaleAndColormap(scale_colmaps=kwargs)
        super().update(**{x: y._asdict() for x, y in wvsc.scale_colmaps.items()})

    def update_from_ini(self, filename):
        cfg = ConfigParser()
        if not os.path.exists(filename):
            raise FileNotFoundError(
                f"VarWebScaleAndColormap update_from_ini('{filename}') which does not exist"
            )
        cfg.read(filename)
        # remove configparser default
        cfg_dict = dict()
        keys = ("scale", "colmap")
        for s in cfg.sections():
            items = []
            for k in keys:
                if k in cfg[s]:
                    if k == "scale":
                        try:
                            items.append(json.loads(cfg[s][k]))
                        except Exception as ex:
                            raise KeyError(
                                f"wrong value for '{k}' of var '{s}' in {filename}: {ex}"
                            )
                    else:
                        items.append(cfg[s][k])
                else:
                    raise KeyError(f"missing '{k}' for var '{s}' in {filename}")
            cfg_dict[s] = tuple(items)
        self.update(**cfg_dict)

    def write(self, filename: str):
        """write the scales and colormaps to a configuration file

        :param filename: path to a filename, currently a .ini file
        """
        cfg = ConfigParser()
        cfg.update(self)
        with open(filename, "w") as fh:
            cfg.write(fh)


#: Mapping of pyaerocom variable names to web naming conventions
class VerticalType(str, Enum):
    """A 2D variable is defined under Column on the website, 3D is defined under Surface"""

    T2D = "2D"
    T3D = "3D"
    UNDEFINED = "UNDEFINED"

    def __str__(self):
        return self.value


class CategoryType(str, Enum):
    optical = "Optical properties"
    particle_conc = "Particle concentrations"
    height = "Height"
    concentration = "Concentration"
    gas_conc = "Gas concentrations"
    vmr = "Volume mixing ratios"
    gas_vmr = "Gas volume mixing ratio"
    deposition = "Deposition"
    temperature = "Temperature"
    particle_ratio = "Particle ratio"
    vertical_column_density = "Vertical column density"
    surface_emission = "Surface emission"
    column_burden = "Column burden"
    solar_irradiance = "Solar irradiance"
    UNDEFINED = "UNDEFINED"

    def __str__(self):
        return self.value


class VariableInfo(NamedTuple):
    menu_name: str
    vertical_type: VerticalType
    category: CategoryType


class _VarWebInfo(BaseModel):
    """Pydantic helper class to ensure the VarWebInfo container always contains the correct data

    :param BaseModel: _description_
    """

    var_web_info: dict[str, VariableInfo]


class VarWebInfo:
    _var_web_info: _VarWebInfo

    def __init__(self, config_file="", **kwargs):
        """This class contains var_web_info and can be accessed like a read-only dict. It
        reads it initial data from data/var_web_info.ini

        :param config_file: filename to additional or updated VariableInfo items, defaults to None
        """
        self._var_web_info = _VarWebInfo(var_web_info=dict())
        with resources.path("pyaerocom.aeroval.data", "var_web_info.ini") as file:
            self.update_from_ini(file)
        if config_file != "":
            self.update_from_ini(config_file)
        self.update(**kwargs)

    def update(self, *args, **kwargs):
        d = copy.deepcopy(self._var_web_info.var_web_info)
        d.update(*args, **kwargs)
        self._var_web_info = _VarWebInfo(var_web_info=d)

    def update_from_ini(self, filename):
        cfg = ConfigParser()
        cfg.read(filename)
        # remove configparser default
        cfg_dict = dict()
        keys = ("menu_name", "vertical_type", "category")
        for s in cfg.sections():
            items = []
            for k in keys:
                if k in cfg[s]:
                    items.append(cfg[s][k])
                else:
                    raise KeyError(f"missing '{k}' for var '{s}' in {filename}")
            cfg_dict[s] = tuple(items)
        self.update(**cfg_dict)

    ## below functions to behave like a read-only dict
    def copy(self):
        return self._var_web_info.var_web_info.copy()

    def __getitem__(self, key):
        return self._var_web_info.var_web_info[key]

    def __len__(self, key):
        return self._var_web_info.var_web_info.__len__()

    def __repr__(self):
        return repr(self._var_web_info.var_web_info)

    def keys(self):
        return self._var_web_info.var_web_info.keys()

    def has_key(self, k):
        return k in self._var_web_info.var_web_info

    def values(self):
        return self._var_web_info.var_web_info.values()

    def items(self):
        return self._var_web_info.var_web_info.items()

    def __cmp__(self, dict_):
        return self._var_web_info.var_web_info.__cmp__(dict_)

    def __contains__(self, item):
        return self._var_web_info.var_web_info.__contains__(item)

    def __iter__(self):
        return self._var_web_info.var_web_info.__iter__()

    def __unicode__(self):
        return self._var_web_info.var_web_info.__unicode__()


#: Default information for statistical parameters
statistics_defaults = {
    "nmb": {
        "name": "NMB",
        "longname": "Normalized Mean Bias",
        "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
        "colmap": "bwr",
        "unit": "%",
        "decimals": 1,
        "forecast": True,
    },
    "mnmb": {
        "name": "MNMB",
        "longname": "Modified Normalized Mean Bias",
        "scale": [-100, -75, -50, -25, 0, 25, 50, 75, 100],
        "colmap": "bwr",
        "unit": "%",
        "decimals": 1,
        "forecast": True,
    },
    "mb": {
        "name": "MB",
        "longname": "Mean Bias",
        "scale": [
            -0.15,
            -0.1,
            -0.05,
            0,
            0.05,
            0.1,
            0.15,
        ],  # factor to be multiplied by range of data
        "colmap": "bwr",
        "unit": "var",
        "decimals": 1,
    },
    "mab": {
        "name": "MAB",
        "longname": "Mean Absolute Bias",
        "scale": [
            0,
            0.025,
            0.05,
            0.075,
            0.1,
            0.125,
            0.15,
        ],  # factor to be multiplied by range of data
        "colmap": "bwr",
        "unit": "var",
        "decimals": 1,
    },
    "R": {
        "name": "R",
        "longname": "Correlation Coefficient",
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
        "forecast": True,
    },
    "R_spearman": {
        "name": "R Spearman",
        "longname": "R Spearman Correlation",
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
        "time_series": True,
    },
    "fge": {
        "name": "FGE",
        "longname": "Fractional Gross Error",
        "scale": [0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2],
        "colmap": "reverseColmap(RdYlGn)",
        "unit": "1",
        "decimals": 2,
        "forecast": True,
    },
    "nrms": {
        "name": "NRMSE",
        "longname": "Normalized Root Mean Square Error",
        "scale": [0, 25, 50, 75, 100, 125, 150, 175, 200],
        "colmap": "Reds",
        "unit": "%",
        "decimals": 1,
        "time_series": True,
    },
    "rms": {
        "name": "RMSE",
        "longname": "Root Mean Square Error",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
        "forecast": True,
    },
    "data_mean": {
        "name": "Mean-Mod",
        "longname": "Model Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
        "time_series": True,
    },
    "refdata_mean": {
        "name": "Mean-Obs",
        "longname": "Observation Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
        "time_series": True,
    },
    "num_valid": {
        "name": "Nb. Obs",
        "longname": "Number of Valid Observations",
        "scale": None,
        "colmap": None,
        "overall_only": True,
        "unit": "1",
        "decimals": 0,
    },
    "num_coords_with_data": {
        "name": "Nb. Stations",
        "longname": "Number of Stations with data",
        "scale": None,
        "colmap": None,
        "overall_only": True,
        "unit": "1",
        "decimals": 0,
    },
}

# Default information for additional statistical parameters
extended_statistics = {
    "R_spatial_mean": {
        "name": "R-Space",
        "longname": "Spatial R computed from yearly averages",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
        "category": "Regional Time Series",
        "time_series": False,
    },
    "R_temporal_median": {
        "name": "R-Temporal",
        "longname": "R temporal median",
        "overall_only": True,
        "scale": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
        "colmap": "RdYlGn",
        "unit": "1",
        "decimals": 2,
        "category": "Regional Time Series",
        "time_series": False,
    },
}

# For adding num statistics
num_statistics = {
    "num_valid": {
        "name": "Nb. Obs",
        "longname": "Number of Valid Observations",
        "scale": None,
        "colmap": None,
        "unit": "1",
        "decimals": 0,
        "overall_only": True,
        "category": "Regional Time Series",
        "time_series": True,
    },
    "num_coords_with_data": {
        "name": "Nb. Stations",
        "longname": "Number of Stations with data",
        "scale": None,
        "colmap": None,
        "unit": "1",
        "decimals": 0,
        "overall_only": True,
        "category": "Regional Time Series",
        "time_series": True,
    },
}

#: Default information about trend display
statistics_trend = {
    "obs/mod_trend": {
        "name": "Obs/Mod-Trends",
        "longname": "Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Regional Time Series",
        "forecast": False,
    },
    "obs_trend": {
        "name": "Obs-Trends",
        "longname": "Observed Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Regional Time Series",
        "forecast": False,
    },
    "mod_trend": {
        "name": "Mod-Trends",
        "longname": "Modelled Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Regional Time Series",
        "forecast": False,
    },
}

statistics_mean_trend = {
    "obs/mod_mean_trend": {
        "name": "Mean Obs/Mod-Trends",
        "longname": "Mean Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
    "obs_mean_trend": {
        "name": "Mean Obs-Trends",
        "longname": "Observed Mean Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
    "mod_mean_trend": {
        "name": "Mean Mod-Trends",
        "longname": "Modelled Mean Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
}

statistics_median_trend = {
    "obs/mod_median_trend": {
        "name": "Median Obs/Mod-Trends",
        "longname": "Median Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
    "obs_median_trend": {
        "name": "Median Obs-Trends",
        "longname": "Observed Median Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
    "mod_median_trend": {
        "name": "Median Mod-Trends",
        "longname": "Modelled Median Trends",
        "scale": [-5.0, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
        "colmap": "bwr",
        "unit": "%/yr",
        "decimals": 1,
        "category": "Individual Stations",
        "time_series": False,
    },
}
# If doing an obs_only experiment, the only statistics which make sense relate just to the observations
statistics_obs_only = {
    "refdata_mean": {
        "name": "Mean-Obs",
        "longname": "Observation Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
    },
}

# For experiments where only model data is interesting, as with proxy drydep
statistics_model_only = {
    "data_mean": {
        "name": "Mean-Mod",
        "longname": "Model Mean",
        "scale": None,
        "colmap": "coolwarm",
        "unit": "1",
        "decimals": 2,
    },
}



from pydantic import (
    BaseModel,
    Field,
)
from typing import Literal


class TimeSetup(BaseModel):
    """
    Time setup options

    Attributes
    ----------
    add_seasons : bool, default True
        if True, seasons will be ['all', 'DJF', 'MAM', 'JJA', 'SON'], if False, just ['all'].
    use_meteorological_seasons : bool, default False
        if True, then statistics are based on the meteorological definition of seasons. This is relevant
        for periods that are a single year. So if :attr:`add_seasons` is True, for a given year ['DJF'] will
        refer to data from Dec of the previous year (if available) and Jan/Feb of the same year, while if
        :attr:`use_meteorological_seasons` is False, it will be based on data from Jan/Feb and December
        of the same year. Similarly, and weather or not :attr:`add_seasons` is True,
        if :attr:`use_meteorological_seasons` is True, ['all'] (whole year) will refer to data from Dec of
        the previous year to Nov of the same year, while if False, it will refer to data from Jan to Dec
        of the same year.
    """

    DEFAULT_FREQS: Literal["monthly", "yearly"] = "monthly"
    SEASONS: list[str] = ["all", "DJF", "MAM", "JJA", "SON"]
    main_freq: str = "monthly"
    freqs: list[str] = ["monthly", "yearly"]
    periods: list[str] = Field(default_factory=list)
    add_seasons: bool = True
    use_meteorological_seasons: bool = False

    def get_seasons(self):
        """
        Get list of seasons to be analysed

        Returns :attr:`SEASONS` if :attr:`add_seasons` it True, else `[
        'all']` (only whole year).

        Returns
        -------
        list
            list of season strings for analysis

        """
        if self.add_seasons:
            return self.SEASONS
        return ["all"]

    def _get_all_period_strings(self):
        """
        Get list of all period strings for evaluation

        Returns
        -------
        list
            list of period / season strings
        """
        output = []
        for per in self.periods:
            for season in self.get_seasons():
                perstr = f"{per}-{season}"
                output.append(perstr)
        return output



"""
Interface for reading EEA AqERep files (formerly known as Airbase data).
"""

from pyaerocom import const
from pyaerocom.io.read_eea_aqerep_base import ReadEEAAQEREPBase


class ReadEEAAQEREP(ReadEEAAQEREPBase):
    """Class for reading EEA AQErep data

    Extended class derived from  low-level base class :class: ReadUngriddedBase
    that contains the main functionality.
    """

    #: Name of the dataset (OBS_ID)
    DATA_ID = const.EEA_NRT_NAME  # change this since we added more vars?

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    #: Eionet offers 2 data revisions
    #: E2a (near real time) and E1a (quality controlled)
    #: this class reads the E2a data for now.
    # But by changing the base path
    # and this constant, it can also read the E1a data set
    DATA_PRODUCT = "E2a"



import abc
import fnmatch
import logging
import sys
from copy import deepcopy
from datetime import datetime
from typing import Any

import numpy as np

from pyaerocom import const
from pyaerocom.exceptions import DataCoverageError, MetaDataError, StationNotFoundError
from pyaerocom.helpers import isnumeric
from pyaerocom.helpers_landsea_masks import get_mask_value, load_region_mask_xr
from pyaerocom.mathutils import in_range
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.units.helpers import get_standard_unit
from pyaerocom.units.units_helpers import get_unit_conversion_fac

logger = logging.getLogger(__name__)


if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class UngriddedDataMetadata(UngriddedDataContainer):
    """Metadata-implementation for UngriddedDataContainer implementing
    metadata, i.e. metadata, data_revision, var_idx/variables, filter_hist and
    _is_vertical_profile. It does not do anything on the data-structure.

    This will only implement the UngriddedDataContainer partially. The data
    part needs to be implemented independently.

    """

    ALLOWED_VERT_COORD_TYPES = ["altitude"]

    def __init__(self):
        self.metadata = {}
        self._data_revision = {}
        self.var_idx = {}
        self.filter_hist = {}
        self._is_vertical_profile = False

    def _copy_metadata_to(self, other: Self) -> Self:
        """
        deepcopy metadata-fields to the other object
        """
        other.metadata = deepcopy(self.metadata)
        other._data_revision = deepcopy(self._data_revision)
        other.var_idx = deepcopy(self.var_idx)
        other.filter_hist = deepcopy(self.filter_hist)
        other.is_vertical_profile = self._is_vertical_profile

    @override
    def get_data_revision(self, data_id) -> str | None:
        """
        Get data revision for a data_id.

        Parameters
        ----------
        data_id : str
            ID of dataset for which revision is to be retrieved

        Raises
        ------
        MetaDataError
            If multiple revisions are found for this dataset.

        Returns
        -------
        latest revision (None if no revision is available).

        """
        if data_id not in self._data_revision:
            rev = None
            for meta in self.metadata.values():
                if meta["data_id"] == data_id:
                    if rev is None:
                        rev = meta["data_revision"]
                    elif not meta["data_revision"] == rev:
                        raise MetaDataError(
                            f"Found different data revisions for dataset {data_id}"
                        )
            self._data_revision[data_id] = rev
        return self._data_revision[data_id]

    @property
    def _first_meta_idx(self):
        """Give the index of the metadata of the first station/meta

        :return: dict
        """
        # First available metadata index
        if self.is_empty:
            raise DataCoverageError("no (meta)-data")
        return next(iter(self.metadata))

    @property
    @override
    def is_vertical_profile(self):
        """Boolean specifying whether is vertical profile"""
        return self._is_vertical_profile

    @is_vertical_profile.setter
    @override
    def is_vertical_profile(self, value):
        """
        Boolean specifying whether is vertical profile.
        Note must be set in ReadUngridded based on the reader
        because the instance of class used during reading is
        not the same as the instance used later in the workflow
        """
        self._is_vertical_profile = value

    def _list_from_metadata(
        self, metafield, undef=None, unique=False, allow_none=True
    ) -> list[Any]:
        """retrieve a station-metadata field as list

        :param metafield: one of the metadata-fields like "longitude" or "instrument"
        :param undef: default value if undefined
        :param unique: remove None and duplicate, defaults to False
        :param allow_none: allow none in unique lists
        :return: list of the metadata-fields values
        """
        ret_vals = []
        for info in self.metadata.values():
            try:
                val = info[metafield]
                if unique:
                    if val not in ret_vals:
                        if allow_none:
                            ret_vals.append(val)
                        elif val is not None:
                            ret_vals.append(val)
                else:
                    ret_vals.append(val)
            except KeyError:
                if not unique:
                    ret_vals.append(undef)
        return ret_vals

    @property
    @override
    def contains_vars(self) -> list[str]:
        """List of all variables in this dataset"""
        return list(self.var_idx)

    @property
    @override
    def contains_datasets(self):
        """List of all datasets in this object"""
        return self._list_from_metadata("data_id", unique=True)

    @property
    @override
    def contains_instruments(self):
        """List of all instruments in this object"""
        return self._list_from_metadata("instrument_name", unique=True, allow_none=False)

    @property
    @override
    def is_empty(self):
        """Boolean specifying whether this object contains data or not"""
        return True if len(self.metadata) == 0 else False

    @property
    @override
    def is_filtered(self):
        """Boolean specifying whether this data object has been filtered

        Note
        ----
        Details about applied filtering can be found in :attr:`filter_hist`
        """
        if len(self.filter_hist) > 0:
            return True
        return False

    @property
    @override
    def longitude(self):
        """Longitudes of stations"""
        return self._list_from_metadata("longitude", undef=np.nan)

    @property
    @override
    def latitude(self):
        """Latitudes of stations"""
        return self._list_from_metadata("latitude", undef=np.nan)

    @property
    @override
    def altitude(self):
        """Altitudes of stations"""
        return self._list_from_metadata("altitude", undef=np.nan)

    @property
    @override
    def station_name(self):
        """Station-names of stations"""
        return self._list_from_metadata("station_name", undef=np.nan)

    @property
    @override
    def unique_station_names(self):
        """List of unique station names"""
        return sorted(set(self.station_name))

    @property
    @override
    def available_meta_keys(self):
        """List of all available metadata keys

        Note
        ----
        This is a list of all metadata keys that exist in this dataset, but
        it does not mean that all of the keys are registered in all metadata
        blocks, especially if the data is merged from different sources with
        different metadata availability
        """
        metakeys = []
        for meta in self.metadata.values():
            for key in meta:
                if key not in metakeys:
                    metakeys.append(key)
        return metakeys

    @property
    @override
    def nonunique_station_names(self):
        """List of station names that occur more than once in metadata"""
        import collections

        lst = self.station_name
        return [item for item, count in collections.Counter(lst).items() if count > 1]

    @override
    def last_filter_applied(self):
        """Returns the last filter that was applied to this dataset

        To see all filters, check out :attr:`filter_hist`
        """
        if not self.is_filtered:
            raise AttributeError("No filters were applied so far")
        return self.filter_hist[max(self.filter_hist)]

    @override
    def find_station_meta_indices(self, station_name_or_pattern, allow_wildcards=True):
        """Find indices of all metadata blocks matching input station name

        You may also use wildcard pattern as input (e.g. *Potenza*)

        Parameters
        ----------
        station_pattern : str
            station name or wildcard pattern
        allow_wildcards : bool
            if True, input station_pattern will be used as wildcard pattern and
            all matches are returned.

        Returns
        -------
        list
           list containing all metadata indices that match the input station
           name or pattern

        Raises
        ------
        StationNotFoundError
            if no such station exists in this data object
        """
        if not allow_wildcards:

            def compare(x, y):
                return fnmatch.fnmatch(x, y)

        else:

            def compare(x, y):
                return x == y

        idx = []
        for i, meta in self.metadata.items():
            if compare(meta["station_name"], station_name_or_pattern):
                idx.append(i)
        if len(idx) == 0:
            raise StationNotFoundError(
                f"No station available in UngriddedData that matches name {station_name_or_pattern}"
            )
        return idx

    @override
    def check_unit(self, var_name, unit=None):
        """Check if variable unit corresponds to AeroCom unit

        Parameters
        ----------
        var_name : str
            variable name for which unit is to be checked
        unit : :obj:`str`, optional
            unit to be checked, if None, AeroCom default unit is used

        Raises
        ------
        MetaDataError
            if unit information is not accessible for input variable name
        """
        if unit is None:
            unit = get_standard_unit(var_name)

        units = []
        for i, meta in self.metadata.items():
            if var_name in meta["var_info"]:
                try:
                    u = meta["var_info"][var_name]["units"]
                    if u not in units:
                        units.append(u)
                except KeyError:
                    add_str = ""
                    if "unit" in meta["var_info"][var_name]:
                        add_str = (
                            "Corresponding var_info dict contains "
                            'attr. "unit", which is deprecated, please '
                            "check corresponding reading routine. "
                        )
                    raise MetaDataError(
                        f"Failed to access unit information for variable {var_name} "
                        f"in metadata block {i}. {add_str}"
                    )
        if len(units) == 0 and str(unit) != "1":
            raise MetaDataError(
                f"Failed to access unit information for variable {var_name}. Expected unit {unit}"
            )
        for u in units:
            if not get_unit_conversion_fac(u, unit, var_name) == 1:
                raise MetaDataError(f"Invalid unit {u} detected (expected {unit})")

    def _add_to_filter_history(self, info):
        """Add info to :attr:`filter_hist`

        Key is current system time string

        Parameter
        ---------
        info
            information to be appended to filter history
        """
        time_str = datetime.now().strftime("%Y%m%d%H%M%S")
        self.filter_hist[int(time_str)] = info

    @property
    @override
    def station_coordinates(self):
        """dictionary with station coordinates

        Returns
        -------
        dict
            dictionary containing station coordinates (latitude, longitude,
            altitude -> values) for all stations (keys) where these parameters
            are accessible.
        """
        d = {"station_name": [], "latitude": [], "longitude": [], "altitude": []}

        for i, meta in self.metadata.items():
            if "station_name" not in meta:
                logger.debug(f"Skipping meta-block {i}: station_name is not defined")
                continue
            elif not all(name in meta for name in const.STANDARD_COORD_NAMES):
                logger.debug(
                    f"Skipping meta-block {i} (station {meta['station_name']}): "
                    f"one or more of the coordinates is not defined"
                )
                continue

            stat = meta["station_name"]

            if stat in d["station_name"]:
                continue
            d["station_name"].append(stat)
            for k in const.STANDARD_COORD_NAMES:
                d[k].append(meta[k])
        return d

    @property
    def countries_available(self):
        """
        Alphabetically sorted list of country names available
        """
        countries = []
        for idx, meta in self.metadata.items():
            try:
                countries.append(meta["country"])
            except Exception:
                logger.warning("No country information in meta block", idx)
        if len(countries) == 0:
            logger.warning(
                "None of the metadata blocks contains "
                "country information. You may want to "
                "run class method check_set_country first "
                "to automatically assign countries."
            )
        return sorted(set(countries))

    def _init_meta_filters(self, **filter_attributes):
        """Init filter dictionary for :func:`apply_filter_meta`

        Parameters
        ----------
        **filter_attributes
            valid meta keywords that are supposed to be filtered and the
            corresponding filter values (or value ranges)
            Only valid meta keywords are considered (e.g. data_id,
            longitude, latitude, altitude, ts_type)

        Returns
        -------
        tuple
            3-element tuple containing

            - dict: string match filters for metakeys \
              (e.g. dict['data_id'] = 'AeronetSunV2Lev2.daily')
            - dict: in-list match filters for metakeys \
              (e.g. dict['station_name'] = ['stat1', 'stat2', 'stat3'])
            - dict: in-range dictionary for metakeys \
              (e.g. dict['longitude'] = [-30, 30])

        """
        # initiate filters that are checked
        valid_keys = self.metadata[self._first_meta_idx]
        str_f = {}
        list_f = {}
        range_f = {}
        val_f = {}
        for key, val in filter_attributes.items():
            if key not in valid_keys:
                raise OSError(
                    f"Invalid input parameter for filtering: {key}. "
                    f"Please choose from {valid_keys}"
                )

            if isinstance(val, str):
                str_f[key] = val
            elif isnumeric(val):
                val_f[key] = val
            elif isinstance(val, list | np.ndarray | tuple):
                if all([isinstance(x, str) for x in val]):
                    list_f[key] = val
                elif len(val) == 2 and all([isnumeric(x) for x in val]):
                    try:
                        low, high = float(val[0]), float(val[1])
                        if not low < high:
                            raise ValueError("First entry needs to be smaller than 2nd")
                        range_f[key] = [low, high]
                    except Exception:
                        list_f[key] = val
                else:
                    list_f[key] = val
        return (str_f, list_f, range_f, val_f)

    @abc.abstractmethod
    def _len_datapoints(self, meta_idx: float | list[float], var: str | list[str]):
        """Get the number of datapoints for meta_idx and var,
        needed internally by _find_meta_matches to calculate the
        total number of new data-size

        Parameters
        ----------
        meta_idx : index or indices of metadata-ids
        var : variable name or list of names

        Returns
        -------
        number of datapoints matching meta_idx and var
        """
        pass

    def _find_meta_matches(self, negate=None, *filters):
        """Find meta matches for input attributes

        Parameters
        ----------
        negate : list or str, optional
            specified meta key(s) provided in `*filters` that are
            supposed to be treated as 'not valid'. E.g. if
            `station_name="bad_site"` is input in `filter_attributes` and if
            `station_name` is listed in `negate`, then all metadata blocks
            containing "bad_site" as station_name will be excluded in output
            data object.
        *filters
            list of filters to be applied

        Returns
        -------
        tuple
            list of metadata indices that match input filter
        """
        if negate is None:
            negate = []
        elif isinstance(negate, str):
            negate = [negate]
        elif not isinstance(negate, list):
            raise ValueError(f"Invalid input for negate {negate}, need list or str or None")
        meta_matches = []
        var_matches = []
        totnum = 0
        for meta_idx, meta in self.metadata.items():
            if self._check_filter_match(meta, negate, *filters):
                meta_matches.append(meta_idx)
                for var in meta["var_info"]:
                    if var in self.ALLOWED_VERT_COORD_TYPES:
                        continue  # altitude is not actually a variable but is stored in var_info like one
                    var_matches.append(var)
        totnum = self._len_datapoints(meta_matches, var_matches)
        return (meta_matches, totnum)

    def _check_str_filter_match(self, meta, negate, str_f):
        # Check string equality for input meta data and filters. Supports
        # wildcard matching
        for metakey, filterval in str_f.items():
            # key does not exist in this specific meta_block
            if metakey not in meta:
                return False
            # check if this key is in negate list (then result will be True
            # for all that do not match the specified filter input value(s))
            neg = metakey in negate

            # actual value of this key in input metadata
            metaval = meta[metakey]

            # check equality of values
            match = metaval == filterval
            if match:  # direct match found
                if neg:  # key is flagged in negate -> no match
                    return False
            else:  # no direct match found
                # check wildcard match
                if "*" in filterval:  # no wildcard in
                    match = fnmatch.fnmatch(metaval, filterval)
                    if neg:
                        if match:
                            return False
                    else:
                        if not match:
                            return False
                elif not neg:  # no match, no wildcard match and not inverted
                    return False
        return True

    def _check_filter_match(self, meta, negate, str_f, list_f, range_f, val_f):
        """Helper method that checks if station meta item matches filters

        Note
        ----
        This method is used in :func:`apply_filter`
        """
        if not self._check_str_filter_match(meta, negate, str_f):
            return False

        for metakey, filterval in list_f.items():
            if metakey not in meta:
                return False
            neg = metakey in negate
            metaval = meta[metakey]
            match = metaval == filterval
            if match:  # lists are identical
                if neg:
                    return False
            else:
                # value in metadata block is different from filter value
                match = metaval in filterval
                if match:
                    if neg:
                        return False
                else:
                    # current metavalue is not equal the filterlist and is also
                    # not contained in the filterlist. However, one or more
                    # entries in the filterlist may be wildcard
                    if isinstance(metaval, str):
                        found = False
                        for entry in filterval:
                            if "*" in entry:
                                match = fnmatch.fnmatch(metaval, entry)
                                if match:
                                    found = True
                                    if neg:
                                        return False
                        if not found and not neg:
                            return False
        # range filter
        for metakey, filterval in range_f.items():
            if metakey not in meta:
                return False
            neg = metakey in negate
            match = in_range(meta[metakey], filterval[0], filterval[1])
            if (neg and match) or (not neg and not match):
                return False

        for metakey, filterval in val_f.items():
            if metakey not in meta:
                return False
            neg = metakey in negate
            match = meta[metakey] == filterval
            if (neg and match) or (not neg and not match):
                return False
        return True

    @override
    def filter_by_meta(self, negate=None, **filter_attributes):
        """Flexible method to filter these data based on input meta specs

        Parameters
        ----------
        negate : list or str, optional
            specified meta key(s) provided via `filter_attributes` that are
            supposed to be treated as 'not valid'. E.g. if
            `station_name="bad_site"` is input in `filter_attributes` and if
            `station_name` is listed in `negate`, then all metadata blocks
            containing "bad_site" as station_name will be excluded in output
            data object.
        **filter_attributes
            valid meta keywords that are supposed to be filtered and the
            corresponding filter values (or value ranges)
            Only valid meta keywords are considered (e.g. data_id,
            longitude, latitude, altitude, ts_type)

        Returns
        -------
        UngriddedData
            filtered ungridded data object

        Raises
        ------
        NotImplementedError
            if attempt variables are supposed to be filtered (not yet possible)
        IOError
            if any of the input keys are not valid meta key

        Example
        -------
        >>> import pyaerocom as pya
        >>> r = pya.io.ReadUngridded(['AeronetSunV3Lev2.daily'], 'od550aer')
        >>> data = r.read()
        >>> data_filtered = data.filter_by_meta(data_id='AeronetSunV3Lev2.daily',
        ...                                     longitude=[-30, 30],
        ...                                     latitude=[20, 70],
        ...                                     altitude=[0, 1000])
        """

        if "variables" in filter_attributes:
            raise NotImplementedError("Cannot yet filter by variables")

        # separate filters by strin, list, etc.
        filters = self._init_meta_filters(**filter_attributes)

        # find all metadata blocks that match the filters
        meta_matches, totnum_new = self._find_meta_matches(
            negate,
            *filters,
        )
        if len(meta_matches) == len(self.metadata):
            logger.info(f"Input filters {filter_attributes} result in unchanged data object")
            return self
        new = self._new_from_meta_blocks(meta_matches, totnum_new)
        time_str = datetime.now().strftime("%Y%m%d%H%M%S")
        new.filter_hist[int(time_str)] = filter_attributes
        return new

    @override
    def filter_by_projection(
        self, projection, xrange: tuple[float, float], yrange: tuple[float, float]
    ):
        """Filter the ungridded data to a horizontal bounding box given by a projection

        :param projection: a function turning projection(lat, lon) -> (x, y)
        :param xrange: x range (min/max included) in the projection plane
        :param yrange: y range (min/max included) in the projection plane
        """
        meta_matches = []
        totnum = 0
        for meta_idx, meta in self.metadata.items():
            lon = meta["longitude"]
            lat = meta["latitude"]
            x, y = projection(lat, lon)

            match_x = in_range(x, xrange[0], xrange[1])
            match_y = in_range(y, yrange[0], yrange[1])

            if match_x and match_y:
                meta_matches.append(meta_idx)
                for var in meta["var_info"]:
                    if var in self.ALLOWED_VERT_COORD_TYPES:
                        continue  # altitude is not actually a variable but is stored in var_info like one
                    try:
                        totnum += len(self.meta_idx[meta_idx][var])
                    except KeyError:
                        logger.debug(
                            f"Ignoring variable {var} in meta block {meta_idx} "
                            f"since no data could be found"
                        )

        if len(meta_matches) == len(self.metadata):
            logger.info("filter_by_projection result in unchanged data object")
            return self
        new = self._new_from_meta_blocks(meta_matches, totnum)
        return new

    @override
    def apply_region_mask(self, region_id=None):
        """
        TODO : Write documentations

        Parameters
        ----------
        region_id : str or list (of strings)
            ID of region or IDs of multiple regions to be combined
        """
        if region_id not in const.HTAP_REGIONS:
            raise ValueError(
                f"Invalid input for region_id: {region_id}, choose from: {const.HTAP_REGIONS}"
            )

        # 1. find matches -> list of meta indices that are in region
        # 2. Get total number of datapoints -> defines shape of output UngriddedData
        # 3. Create

        mask = load_region_mask_xr(region_id)

        meta_matches = []
        totnum = 0
        for meta_idx, meta in self.metadata.items():
            lon, lat = meta["longitude"], meta["latitude"]

            mask_val = get_mask_value(lat, lon, mask)
            if mask_val >= 1:  # coordinate is in mask
                meta_matches.append(meta_idx)
                for var in meta["var_info"]:
                    totnum += len(self.meta_idx[meta_idx][var])

        new = self._new_from_meta_blocks(meta_matches, totnum)
        time_str = datetime.now().strftime("%Y%m%d%H%M%S")
        new.filter_hist[int(time_str)] = f"Applied mask {region_id}"
        new._check_index()
        return new

    def _meta_to_lists(self):
        """List metadata for plotting

        :return: list of metadata values
        """
        meta = {k: [] for k in self.metadata[self._first_meta_idx]}
        for meta_item in self.metadata.values():
            for k, v in meta.items():
                v.append(meta_item[k])
        return meta

    @override
    def extract_dataset(self, data_id):
        """Extract single dataset into new instance of :class:`UngriddedData`

        Calls :func:`filter_by_meta`.

        Parameters
        -----------
        data_id : str
            ID of dataset

        Returns
        -------
        UngriddedData
            new instance of ungridded data containing only data from specified
            input network
        """
        logger.info(f"Extracting dataset {data_id} from data object")
        return self.filter_by_meta(data_id=data_id)



from __future__ import annotations

import logging
import re
from collections.abc import Iterator
from datetime import date, datetime, timedelta
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm

from pyaerocom.griddeddata import GriddedData
from pyaerocom.io.cams2_83.models import ModelData, ModelName, RunType
from pyaerocom.io.gridded_reader import GriddedReader

"""
TODO:

As it is now, with e.g. leap = 3 and start date 01.12, 01.12 might not be used, since the leap shifts the date three days forward
This might have to be componsated for, with the filepath being for 3 days before 01.12 (the start date)(?)
"""

AEROCOM_NAMES = dict(
    co_conc="concco",
    no2_conc="concno2",
    o3_conc="conco3",
    pm10_conc="concpm10",
    pm2p5_conc="concpm25",
    so2_conc="concso2",
)

FULL_NAMES = dict(
    co_conc="Carbon Monoxide",
    no2_conc="Nitrogen Dioxide",
    o3_conc="Ozone",
    pm10_conc="PM10 Aerosol",
    pm2p5_conc="PM2.5 Aerosol",
    so2_conc="Sulphur Dioxide",
)

STANDARD_NAMES = dict(
    co_conc="mass_concentration_of_carbon_monoxide_in_air",
    no2_conc="mass_concentration_of_nitrogen_dioxide_in_air",
    o3_conc="mass_concentration_of_ozone_in_air",
    pm10_conc="mass_concentration_of_pm10_ambient_aerosol_in_air",
    pm2p5_conc="mass_concentration_of_pm2p5_ambient_aerosol_in_air",
    so2_conc="mass_concentration_of_sulfur_dioxide_in_air",
)


DATA_FOLDER_PATH = Path("/lustre/storeB/project/fou/kl/CAMS2_83/model")


DEBUG = True


logger = logging.getLogger(__name__)


def __model_path(
    name: str | ModelName,
    date: str | date | datetime,
    *,
    root_path: Path | str,
    run: str | RunType,
) -> Path:
    if not isinstance(name, ModelName):
        name = ModelName[name]
    if isinstance(date, str):
        date = datetime.strptime(date, "%Y%m%d").date()
    if isinstance(date, datetime):
        date = date.date()
    if isinstance(root_path, str):
        root_path = Path(root_path)
    if not isinstance(run, RunType):
        run = RunType[run]
    return ModelData(name, run, date, root_path).path


def model_paths(
    model: str | ModelName,
    *dates: datetime | date | str,
    root_path: Path | str = DATA_FOLDER_PATH,
    run: str | RunType = RunType.FC,
) -> Iterator[Path]:
    for date in dates:  # noqa: F402
        path = __model_path(model, date, run=run, root_path=root_path)
        if not path.is_file():
            logger.warning(f"Could not find {path.name}. Skipping {date}")
            continue
        yield path


def parse_daterange(
    dates: pd.DatetimeIndex | list[datetime] | tuple[datetime, datetime],
) -> pd.DatetimeIndex:
    if isinstance(dates, pd.DatetimeIndex):
        return dates
    if len(dates) != 2:
        raise ValueError("need 2 datetime objects to define a date_range")
    return pd.date_range(*dates, freq="d")


def forecast_day(ds: xr.Dataset, *, day: int) -> xr.Dataset:
    data = ModelData.frompath(ds.encoding["source"])
    if not (0 <= day <= data.run.days):
        raise ValueError(f"{data} has no day #{day}")
    date = data.date + timedelta(days=day)
    first_date = data.date

    select_date = datetime(date.year, date.month, date.day, 0, 0, 0)

    dateselect = pd.date_range(select_date, select_date + timedelta(hours=23), freq="h")

    if isinstance(ds.time.data[0], np.timedelta64):
        logger.debug(f"Changing time dimension for {ds.encoding['source']}")
        ds = ds.assign_coords(time=np.datetime64(first_date) + ds.time.data)

    if len(set(np.array(ds.time))) != len(np.array(ds.time)):
        logger.debug(f"Changing time dimension for {ds.encoding['source']}")
        ds = ds.assign_coords(time=dateselect)

    # if len(ds.time.data) < 2:
    #     return xr.Dataset()
    try:
        ds = ds.sel(time=dateselect)
    except Exception as e:
        logger.debug(f"{e}")
        logger.debug(f"Interpolating NaNs for {ds.encoding['source']}")
        ds = ds.interp(time=dateselect)
        ds = ds.sel(time=dateselect)

    ds = ds.sel(level=0.0)
    ds.time.attrs["long_name"] = "time"
    ds.time.attrs["standard_name"] = "time"

    for var_name in ds.data_vars:
        ds[var_name].attrs["forecast_day"] = day
    return ds


def fix_coord(ds: xr.Dataset) -> xr.Dataset:
    lon = ds.longitude.data
    ds["longitude"] = np.where(lon > 180, lon - 360, lon)
    ds.longitude.attrs.update(
        long_name="longitude", standard_name="longitude", units="degrees_east"
    )
    ds.latitude.attrs.update(long_name="latitude", standard_name="latitude", units="degrees_north")
    return ds


def fix_names(ds: xr.Dataset) -> xr.Dataset:
    for var_name, aerocom_name in AEROCOM_NAMES.items():
        ds[var_name].attrs.update(long_name=aerocom_name)
    return ds.rename(AEROCOM_NAMES)


def fix_missing_vars(ds: xr.Dataset) -> xr.Dataset:
    """
    TODO: Check if all variables are there. If not:
    make the rest of the variables, filled with nans.
    Log an error when this is done

    Might not be possible...
    """
    vars_list = [i for i in ds.data_vars]
    nb_vars = len(vars_list)
    if nb_vars < 6:
        logger.warning(f"Found only {vars_list}. Filling the rest with NaNs")

        dummy_var = ds[vars_list[0]]
        dummy_var_name = vars_list[0]
        for species in AEROCOM_NAMES:
            if species not in vars_list:
                ds = ds.assign(**{species: dummy_var * np.nan})
                attrs = ds[dummy_var_name].attrs
                attrs["species"] = FULL_NAMES[species]
                attrs["standard_name"] = STANDARD_NAMES[species]
                ds[species] = ds[species].assign_attrs(attrs)
    return ds


def read_dataset(paths: list[Path], *, day: int) -> xr.Dataset:
    paths = check_files(paths)

    def preprocess(ds: xr.Dataset) -> xr.Dataset:
        return ds.pipe(forecast_day, day=day).pipe(fix_missing_vars)

    ds = xr.open_mfdataset(paths, preprocess=preprocess, parallel=False, chunks={"time": 24})
    return ds.pipe(fix_coord).pipe(fix_names)


def check_files(paths: list[Path]) -> list[Path]:
    if not DEBUG:
        return paths

    new_paths: list[Path] = []

    for p in tqdm(paths, disable=None):
        try:
            with xr.open_dataset(p, decode_timedelta=True) as ds:
                if len(ds.time.data) < 2:
                    logger.warning(f"To few timestamps in {p}. Skipping file")
                    continue
                if len(set(np.array(ds.time))) != len(np.array(ds.time)):
                    if len(np.array(ds.time)) != 24:
                        logger.warning(
                            f"Ambiguous time dimension: Duplicate timestamps in {p}, with less that 24 step. Skipping file"
                        )
                        continue

            new_paths.append(p)
        except Exception as ex:
            logger.warning(f"Error when opening {p}: {ex}. Skipping file")

    return new_paths


class ReadCAMS2_83(GriddedReader):
    FREQ_CODES = dict(hour="hourly")
    REVERSE_FREQ_CODES = {val: key for key, val in FREQ_CODES.items()}

    def __init__(
        self,
        data_id: str | None = None,
        data_dir: str | Path | None = None,
    ) -> None:
        self._filedata: xr.Dataset | None = None
        self._filepaths: list[Path] | None = None
        self._data_dir: Path | None = None
        self._model: ModelName | None = None
        self._forecast_day: int | None = None
        self._data_id: str | None = None
        self._daterange: pd.DatetimeIndex | None = None

        self._run_type: RunType = RunType.FC

        if data_dir is not None:
            if isinstance(data_dir, str):
                data_dir = Path(data_dir)
            self.data_dir = data_dir

        if data_id is not None:
            self.data_id = data_id

    @property
    def data_dir(self) -> Path:
        """
        Directory containing netcdf files
        """
        if self._data_dir is None:
            raise AttributeError("data_dir needs to be set before accessing")
        return self._data_dir

    @data_dir.setter
    def data_dir(self, val: str | Path | None):
        if val is None:
            raise ValueError(f"Data dir {val} needs to be a dictionary or a file")
        if isinstance(val, str):
            val = Path(val)
        if not val.is_dir():
            raise NotADirectoryError(val)
        self._data_dir = val
        self._filedata = None

    @property
    def data_id(self):
        if self._data_id is None:
            raise AttributeError("data_id needs to be set before accessing")
        return self._data_id

    @data_id.setter
    def data_id(self, val):
        if val is None:
            raise ValueError(f"The data_id {val} can't be None")
        elif not isinstance(val, str):
            raise TypeError(f"The data_id {val} needs to be a string")

        self._data_id = val

        match = re.match(r"^CAMS2-83\.(.*)\.day(\d)\.(FC|AN)$", val)
        if match is None:
            raise ValueError(f"The id {val} is not on the correct format")

        model, day, run_type = match.groups()
        self.run_type = RunType[run_type]
        self.model = ModelName[model]
        self.forecast_day = int(day)

    @property
    def years_avail(self):
        return np.unique(
            self.daterange.values.astype("datetime64[Y]").astype("int") + 1970
        ).astype("str")

    @property
    def ts_types(self):
        return self.REVERSE_FREQ_CODES.keys()

    @property
    def vars_provided(self):
        return AEROCOM_NAMES.values()

    @property
    def run_type(self):
        if self._run_type is None:
            raise AttributeError("run_type needs to be set before accessing")
        return self._run_type

    @run_type.setter
    def run_type(self, val):
        if val is None:
            raise AttributeError("run_type cannot be set as None")
        elif not isinstance(val, RunType):
            raise AttributeError(f"run_type cannot be set as {type(val)}, but must be a RunType")

        self._run_type = val

    @property
    def filepaths(self) -> list[Path]:
        """
        Path to data file
        """
        if self.data_dir is None and self._filepaths is None:  # type:ignore[unreachable]
            raise AttributeError("data_dir or filepaths needs to be set before accessing")
        if self._filepaths is None:
            paths = list(
                model_paths(
                    self.model,
                    *self.daterange,
                    root_path=self.data_dir,
                    run=self.run_type,
                )
            )
            if not paths:
                raise ValueError(f"no files found for {self.model}")
            self._filepaths = paths
        return self._filepaths

    @filepaths.setter
    def filepaths(self, value: list[Path]):
        if not bool(list):
            raise ValueError("needs to be list of paths")
        if not isinstance(value, list):
            raise ValueError("needs to be list of paths")
        if all(isinstance(path, Path) for path in value):
            raise ValueError("needs to be list of paths")
        self._filepaths = value

    @property
    def filedata(self) -> xr.Dataset:
        """
        Loaded netcdf file (:class:`xarray.Dataset`)
        """
        if self._filedata is None:
            self._filedata = read_dataset(self.filepaths, day=self.forecast_day)
        return self._filedata

    @property
    def model(self) -> str:
        if self._model is None:
            raise ValueError("Model not set")
        return self._model

    @model.setter
    def model(self, val: str | ModelName):
        if not isinstance(val, ModelName):
            val = ModelName(val)
        self._model = val
        self._filedata = None

    @property
    def daterange(self) -> pd.DatetimeIndex:
        if self._daterange is None:
            raise ValueError("The date range is not set yet")
        return self._daterange

    @daterange.setter
    def daterange(self, dates: pd.DatetimeIndex | list[datetime] | tuple[datetime]):
        if not isinstance(dates, pd.DatetimeIndex | list | tuple):
            raise TypeError(f"{dates} need to be a pandas DatetimeIndex or 2 datetimes")

        self._daterange = parse_daterange(dates)
        self._filedata = None

    @property
    def forecast_day(self) -> int:
        if self._forecast_day is None:
            raise ValueError("forecast_day is not set")
        return self._forecast_day

    @forecast_day.setter
    def forecast_day(self, val: int):
        if not isinstance(val, int) or not (0 <= val <= 3):
            raise TypeError(f"forecast_day {val} is not a int between 0 and 3")
        self._forecast_day = val

    @staticmethod
    def has_var(var_name):
        """Check if variable is supported

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        return var_name in AEROCOM_NAMES.values()

    def read_var(self, var_name: str, ts_type: str | None = None, **kwargs) -> GriddedData:
        """Load data for given variable.

        Parameters
        ----------
        var_name : str
            Variable to be read
        ts_type : str
            Temporal resolution of data to read. Supported are
            "hourly", "daily", "monthly" , "yearly".

        Returns
        -------
        GriddedData
        """
        if "daterange" in kwargs:
            self.daterange = kwargs["daterange"]
        if self._daterange is None:
            raise ValueError(f"No 'daterange' in kwargs={kwargs}")

        if ts_type != "hourly":
            raise ValueError("Only hourly ts_type is supported")

        cube = self.filedata[var_name].to_iris()

        gridded = GriddedData(
            cube,
            var_name=var_name,
            ts_type=ts_type,
            check_unit=True,
            convert_unit_on_init=True,
        )
        gridded.metadata["data_id"] = self.data_id
        return gridded


if __name__ == "__main__":
    # from time import perf_counter

    data_dir = str(DATA_FOLDER_PATH)
    data_id = "CAMS2-83.EMEP.day0.AN"
    reader = ReadCAMS2_83(data_dir=data_dir, data_id=data_id)
    reader.daterange = ("2021-12-01", "2021-12-04")
    print(
        np.unique(reader.daterange.values.astype("datetime64[Y]").astype("int") + 1970).astype(
            "str"
        )
    )
    print(reader.filepaths)
    # dates = ("2021-12-01", "2021-12-04")

    # seconds = -perf_counter()
    # print(reader.read_var("concno2", ts_type="hourly", daterange=dates))

    # seconds += perf_counter()
    # print(timedelta(seconds=int(seconds)))



"""
Global config for emep reporting pyaeroval runs
"""

import copy
import functools
import logging
import os

import yaml

from pyaerocom.data import resources

logger = logging.getLogger(__name__)


# Constraints
DEFAULT_RESAMPLE_CONSTRAINTS = dict(
    yearly=dict(monthly=9),
    monthly=dict(
        daily=21,
        weekly=3,
    ),
    daily=dict(hourly=18),
)

DEFAULT_RESAMPLE_CONSTRAINTS_DAILY = dict(
    daily=dict(hourly=18),
)

OC_EC_RESAMPLE_CONSTRAINTS = dict(
    yearly=dict(monthly=4),
    monthly=dict(daily=4, weekly=1),
    daily=dict(hourly=18),
    hourly=dict(minutely=45),
)


OC_EC_RESAMPLE_CONSTRAINTS_DAILY = dict(
    # monthly=dict(daily=4, weekly=1),
    daily=dict(hourly=18),
    hourly=dict(minutely=45),
)


@functools.cache
def _get_ignore_stations_from_file():
    if os.path.exists("./omit_stations.yaml"):
        filename = os.path.abspath("./omit_stations.yaml")
        logger.info(f"reading omit_stations.yaml from {filename}")
        with open(filename) as fh:
            stations = yaml.safe_load(fh)
    else:
        with resources.path(__package__, "omit_stations.yaml") as filename:
            logger.info(f"reading omit_stations.yaml from {filename}")
            with filename.open() as fh:
                stations = yaml.safe_load(fh)

    rows = []
    for year, comps in stations.items():
        if year == "variables":
            continue
        year = int(year)
        for comp, stats in comps.items():
            for stat in stats:
                for var in stations["variables"][comp]:
                    rows.append((year, year, var.strip(), stat.strip()))
    return rows


def _get_ignore_stations(specy, year):
    """
    Read the ignore stations from either omit_stations.tsv in the local eller in the lib-folder

    specy: specy for this measurement network (ALL are translated to all specy)
    year: only select the stations for the specified year

    return: list of stations
    """
    retvals = []
    year = int(year)
    stations = _get_ignore_stations_from_file()
    for yearstart, yearend, comp, station in stations:
        if comp == "ALL" or comp == specy:
            if yearstart <= year <= yearend:
                retvals.append(station)
    return retvals


def get_CFG(reportyear, year, model_dir) -> dict:
    """Get a configuration usable for emep reporting

    :param reportyear: year of reporting
    :param year: year of data
    :param model_dir: directory containing Base_hour.nc,Base_day.nc,Base_month.nc and Base_fullrun.nc
        or for trends directory containing years like 2005,2010,2015 again containing above files

    The current working directory of the experiment should have the following files/directories by default:
        - `data` output directory
        - `coldata` output directory
        - `user_var_scale_colmap.ini` optional user-defined colormaps for pyaerocom variables
        - `omit_stations.yaml` optional user-defined yaml file of stations to omit

    The default values can be changed in your program. If you want to permanently change the defaults,
    please agree upon these changes with the emep-modellers and contact the pyaerocom-developers.

    Example runs with this config look like::

        import os
        import pyaerocom as pya
        from pyaerocom import const
        from pyaerocom.aeroval import EvalSetup, ExperimentProcessor

        from pyaerocom.aeroval.config.emep.reporting_base import get_CFG

        # Setup for models used in analysis
        CFG = get_CFG(reportyear=2024,
                    year=2021,
                    model_dir="/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_metyear2021_emis2022")

        CFG.update(dict(
            # proj_id="status-2024",
            exp_id="test-2021met_2022emis",
            exp_name="Test runs for 2024 EMEP reporting",
            exp_descr=(
                "Test run from Agnes for 2024_REPORTING/EMEP01_rv5.3_metyear2021_emis2022, i.e. 2021met and 2022emis"
            ),
            exp_pi="S. Tsyro, A. Nyiri, H. Klein",
        ))

        # remove EEA
        # for obs in list(CFG["obs_cfg"].keys()):
        #     if obs.startswith("EEA"):
        #         del CFG["obs_cfg"][obs]
        #         print(f"removed {obs}")

        # remove "concCocpm10", not in model-output
        for obs in CFG["obs_cfg"]:
            if "concCocpm10" in CFG["obs_cfg"][obs]["obs_vars"]:
                CFG["obs_cfg"][obs]["obs_vars"].remove("concCocpm10")

        # remove "no, pm10, pm25" from EBAS-hourly
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concNno")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm10")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm25")


        # CFG["raise_exceptions"] = False
        # CFG["add_model_maps"] = False
        # CFG["only_model_maps"] = True


        stp = EvalSetup(**CFG)
        cdir = "./cache/"
        os.makedirs(cdir, exist_ok=True)
        const.CACHEDIR = cdir

        ana = ExperimentProcessor(stp)
        ana.update_interface()

        res = ana.run()

    Another example for multiple model-evaluation::

        import os
        import pyaerocom as pya
        from pyaerocom import const
        from pyaerocom.aeroval import EvalSetup, ExperimentProcessor

        from pyaerocom.aeroval.config.emep.reporting_base import get_CFG

        # Setup for models used in analysis
        CFG = get_CFG(
            reportyear=2024,
            year=2022,
            model_dir=f"/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_year2022_Status_Rep2024",
        )

        dir_versions = {
            "FFmod": "/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_year2022_Status_Rep2024_FFmod/",
            "MARS5.3": "/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_year2022_Status_Rep2024_MARS/",
            "MARS5.0": "/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.0_year2022_Status_Rep2023_emis2022/",
            "NoCations": "/lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.3_year2022_Status_Rep2024_noCation/",
        }

        # Comparison of several models
        MODEL = CFG["model_cfg"]["EMEP"]
        PLTTYPES = CFG["plot_types"]["EMEP"]
        for mid, fpath in dir_versions.items():
            CFG["model_cfg"][mid] = MODEL.copy()
            CFG["plot_types"][mid] = PLTTYPES.copy()
            CFG["model_cfg"][mid]["model_data_dir"] = fpath
            CFG["model_cfg"][mid]["model_id"] = mid
        del CFG["model_cfg"]["EMEP"]
        del CFG["plot_types"]["EMEP"]

        # change some config settings, usually not needed
        CFG.update(
            dict(
                proj_id="emepX",
                exp_id=f"2024-XXX_2022_ebas2",
                # exp_name="Evaluation of EMEP runs for 2023 EMEP reporting",
                exp_descr=(
                    f"Evaluation of EMEP runs for 2024 EMEP reporting, MARS vs ISOROPIA. /lustre/storeB/project/fou/kl/emep/ModelRuns/2024_REPORTING/EMEP01_rv5.?_year2022_Status_Rep2024_*/, is compared against observations from EBAS."
                ),
                # periods=["2021"],
                # exp_pi="S. Tsyro, H. Klein",
                # add_model_maps=False,
            )
        )

        # remove "concCocpm10", not in model-output
        for obs in CFG["obs_cfg"]:
            if "concCocpm10" in CFG["obs_cfg"][obs]["obs_vars"]:
                CFG["obs_cfg"][obs]["obs_vars"].remove("concCocpm10")

        # remove "no, pm10, pm25" from EBAS-hourly
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concNno")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm10")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm25")

        # remove EEA
        for obs in list(CFG["obs_cfg"].keys()):
            if obs.startswith("EEA"):
                del CFG["obs_cfg"][obs]
                print(f"removed {obs}")


        # try to run anything, but don't fail on error
        # CFG["raise_exceptions"] = False


        stp = EvalSetup(**CFG)

        cdir = "./cache"
        os.makedirs(cdir, exist_ok=True)
        const.CACHEDIR = cdir

        ana = ExperimentProcessor(stp)
        ana.update_interface()

        # run everything
        res = ana.run()

    and the example for trends::

        import os
        import pyaerocom as pya
        from pyaerocom import const
        from pyaerocom.aeroval import EvalSetup, ExperimentProcessor

        from pyaerocom.aeroval.config.emep.reporting_base import get_CFG

        # Setup for models used in analysis
        CFG = get_CFG(reportyear=2023,
                    year=2021,
                    model_dir=f"/lustre/storeB/project/fou/kl/emep/ModelRuns/2023_REPORTING/TRENDS/pyaerocom_trends/")


        CFG.update(dict(
            proj_id="emep",
            exp_id=f"2023-trends",
            # exp_name="Evaluation of EMEP runs for 2023 EMEP reporting",
            exp_descr=(
                f"Evaluation of EMEP runs for 2023 EMEP reporting trend runs. 7 year obs-data availability per period. /lustre/storeB/project/fou/kl/emep/ModelRuns/2023_REPORTING/TRENDS/pyaerocom_trends is compared against observations fro
        m EBAS."
            ),
            periods=["1990-2021", "1990-1999", "2000-2009", "2010-2019", "2012-2021"], #range(1990,2022)],
            # exp_pi="S. Tsyro, H. Klein",
            add_model_maps=False,
            #only_model_maps=True,
            # trend parameters
            freqs=["yearly", "monthly"], # "weekly"],"daily"], # can't be hourly for trends, daily is too slow weekly hardly ever needed
            main_freq="monthly",
            add_trends=True,
            avg_over_trends=True,
            obs_min_yrs=7, # kun stasjoner med minst 14yr
            stats_min_yrs=7, # kun stasjoner med minst 14yr
            sequential_yrs=False,
        ))


        # remove "no, pm10, pm25" from EBAS-hourly
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concNno")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm10")
        CFG["obs_cfg"]["EBAS-h-diurnal"]["obs_vars"].remove("concpm25")


        # remove EEA
        for obs in list(CFG["obs_cfg"].keys()):
            if obs.startswith("EEA"):
                del CFG["obs_cfg"][obs]

        # remove all hourly obs, f.e. for trends
        for obs in list(CFG["obs_cfg"].keys()):
            if "ts_type" in CFG["obs_cfg"][obs] and CFG["obs_cfg"][obs]["ts_type"] == "hourly":
                del CFG["obs_cfg"][obs]
                print(f"removed hourly {obs}")

        # remove all daily obs, f.e. for trends
        for obs in list(CFG["obs_cfg"].keys()):
            if "ts_type" in CFG["obs_cfg"][obs] and CFG["obs_cfg"][obs]["ts_type"] == "daily":
                del CFG["obs_cfg"][obs]
                print(f"removed daily {obs}")


        # remove "concCocpm10", not in model-output
        for obs in CFG["obs_cfg"]:
            if "concCocpm10" in CFG["obs_cfg"][obs]["obs_vars"]:
                CFG["obs_cfg"][obs]["obs_vars"].remove("concCocpm10")

                # try to run anything, but don't fail on error
        # CFG["raise_exceptions"] = False


        stp = EvalSetup(**CFG)

        cdir = "./cache"
        os.makedirs(cdir, exist_ok=True)
        const.CACHEDIR = cdir

        ana = ExperimentProcessor(stp)
        ana.update_interface()

        # run everything
        res = ana.run()

    :returns: a dict of a model configuration usable for EvalSetup
    """

    CFG = dict(
        json_basedir=os.path.abspath("./data"),
        coldata_basedir=os.path.abspath("./coldata"),
        # io_aux_file=os.path.abspath("./gridded_io_aux.py"), not needed for ReadMscwCtm
        var_scale_colmap_file=os.path.abspath("./user_var_scale_colmap.ini"),
        # if True, existing colocated data files will be deleted and contours will be overwritten
        reanalyse_existing=True,
        only_json=False,
        add_model_maps=True,
        only_model_maps=False,
        boundaries={
            "west": -30,
            "east": 90,
            "north": 82,
            "south": 30,
        },
        maps_freq="yearly",
        plot_types={"EMEP": ["contour", "overlay"]},
        clear_existing_json=False,
        # if True, the analysis will stop whenever an error occurs (else, errors that
        # occurred will be written into the logfiles)
        raise_exceptions=True,
        # Regional filter for analysis
        filter_name="ALL-wMOUNTAINS",
        # colocation frequency (no statistics in higher resolution can be computed)
        ts_type="daily",
        map_zoom="Europe",
        freqs=["yearly", "monthly", "weekly", "daily", "hourly"],
        periods=[f"{year}"],
        main_freq="daily",
        zeros_to_nan=False,
        use_diurnal=True,
        min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
        colocate_time=True,
        resample_how={"vmro3max": {"daily": {"hourly": "max"}}},
        obs_remove_outliers=False,
        model_remove_outliers=False,
        harmonise_units=True,
        regions_how="country",
        annual_stats_constrained=True,
        proj_id="emep",
        exp_id=f"{reportyear}-reporting",
        exp_name=f"Evaluation of EMEP runs for {reportyear} EMEP reporting",
        exp_descr=(
            f"Evaluation of EMEP runs for {reportyear} EMEP reporting. The EMEP model, simulated for {year}, is compared against observations from EEA and EBAS."
        ),
        exp_pi="emep.mscw@met.no",
        public=True,
        # directory where colocated data files are supposed to be stored
        weighted_stats=True,
        var_order_menu=[
            # Gases
            "concNno",
            "concNno2",
            "concNtno3",
            "concNhno3",
            "concNtnh",
            "concNnh3",
            "concnh4",
            "concSso2",
            "concso4t",
            "concso4c",
            "vmro3",
            "vmro3max",
            "vmro3mda8",
            "vmrox",
            "vmrco",
            # PMs
            "concpm10",
            "concpm25",
            "concno3pm10",
            "concno3pm25",
            "concno3pm1",
            "concnh4pm25",
            "concnh4pm1",
            "concso4pm25",
            "concso4pm1",
            "concCecpm10",
            "concCecpm25",
            "concCocpm10",  # SURF_ugC_PM_OMCOARSE missing in model-output
            "concCocpm25",
            "concom1",
            "concsspm10",
            "concsspm25",
            # Depositions
            "wetrdn",
            "wetoxs",
            "wetoxn",
            "prmm",
        ],
    )

    CFG["model_cfg"] = {
        "EMEP": dict(
            model_id="EMEP",
            model_data_dir=model_dir,
            gridded_reader_id={"model": "ReadMscwCtm"},
            model_read_aux={},
            # model_ts_type_read="daily",
        ),
    }

    """
    Filters
    """

    # OBS SPECIFIC FILTERS (combination of the above and more)
    BASE_FILTER = {
        "latitude": [30, 82],
        "longitude": [-30, 90],
    }

    EBAS_FILTER = {
        **BASE_FILTER,
        "data_level": [None, 2],
        "set_flags_nan": True,
    }

    AERONET_FILTER = {
        **BASE_FILTER,  # Forandring fra Daniel
        "altitude": [-20, 1000],
    }

    # Station filters

    ebas_species = [
        "concNhno3",
        "concNtno3",
        "concNtnh",
        "concNnh3",
        "concnh4",
        "prmm",
        "concpm10",
        "concpm25",
        "concSso2",
        "concNno2",
        "vmrco",
        "vmro3max",
        "vmro3",
        "concNno",
        "concCecpm25",
        "concCocpm25",
        "concom1",
        "concCecpm10",
        "concCocpm10",
        #        "concnh4pm10", # no output in the model
        "concnh4pm25",
        "concnh4pm1",
        #        "concso4pm10", # no output in the model
        "concso4pm25",
        "concso4pm1",
        "concno3pm10",
        "concno3pm25",
        "concno3pm1",
        "concsspm10",
        "concsspm25",
        "concso4t",
        "concso4c",
        "wetoxs",
        "wetoxn",
        "wetrdn",
        "vmrox",
    ]

    # This list of stations was generated using the script found here:
    # https://gist.github.com/thorbjoernl/b7946882f1696722742053406d056e12.
    # It excludes stations with a relative altitude (Elevation difference to the lowest
    # altitude in a 5km radius based on gtopo30) above 500m as well as stations that do not include
    # an altitude in the ebas file index.
    # Last updated: ~2025-01-03
    height_ignore_ebas = [
        "AM0001R",
        "AR0001R",
        "AT0033R",
        "AT0034G",
        "AT0037R",
        "AT0038R",
        "AT0040R",
        "AT0048R",
        "AT0049R",
        "BG0001R",
        "BG0053R",
        "BO0001R",
        "CA0100R",
        "CA0103R",
        "CH0001G",
        "CH0004R",
        "CH0005R",
        "CL0001R",
        "CN1003R",
        "DE0003R",
        "DE0005R",
        "DE0054R",
        "DE0057G",
        "DE0060G",
        "DE0075R",
        "DZ0001G",
        "ES0005R",
        "ES0018G",
        "ES0022R",
        "ES0025U",
        "FI0009R",
        "FR0012R",
        "FR0019R",
        "FR0026R",
        "FR0030R",
        "FR0031R",
        "FR0033R",
        "GB0035R",
        "GB0059G",
        "GR0003R",
        "GR0101R",
        "HR0002R",
        "HR0004R",
        "IT0002R",
        "IT0003R",
        "IT0005R",
        "IT0009R",
        "IT0019R",
        "IT0031U",
        "JP1021R",
        "KE0001G",
        "MK0007R",
        "MX0001R",
        "MY1030R",
        "NO0036R",
        "PL0003R",
        "PL0011R",
        "PT0005R",
        "PT0007R",
        "RO0001R",
        "RO0002R",
        "RO0003R",
        "RO0004R",
        "RO0005R",
        "RS0005R",
        "RU1038R",
        "SI0032R",
        "SK0002R",
        "TW0100R",
        "US0012R",
        "US0013R",
        "US0015R",
        "US0016R",
        "US0024R",
        "US0030R",
        "US0032R",
        "US0053R",
        "US0054R",
        "US0055R",
        "US0073R",
        "US0077R",
        "US0082R",
        "US0131R",
        "US0142R",
        "US0204R",
        "US0602R",
        "US1200R",
        "US4828R",
        "US9002R",
        "US9005R",
        "US9020R",
        "US9024R",
        "US9026R",
        "US9029R",
        "US9041R",
        "US9042R",
        "US9046R",
        "US9048R",
        "US9050R",
        "US9056R",
        "US9064R",
        "US9065R",
        "US9070R",
        "US9071R",
        "US9078R",
        "US9082U",
        "VN0001R",
    ]

    EBAS_FILTER = {
        key: dict(
            **EBAS_FILTER,
            station_id=_get_ignore_stations(key, year) + height_ignore_ebas,
            negate="station_id",
        )
        for key in ebas_species
    }

    EEA_FILTER = {
        **BASE_FILTER,
    }

    OBS_GROUNDBASED = {
        ##################
        #    EBAS
        ##################
        "EBAS-m-tc": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-m",
            obs_vars=[
                "concNhno3",
                "concNtno3",
                "concNtnh",
                "concNnh3",
                "concnh4",
                # "prmm",
                "concpm10",
                "concpm25",
                "concSso2",
                "concNno2",
                "vmrco",
                "vmro3max",
                "vmro3",
                "concNno",
                "concso4t",
                "concso4c",
            ],
            obs_vert_type="Surface",
            colocate_time=True,
            ts_type="monthly",
            obs_filters=EBAS_FILTER,
        ),
        "EBAS-d-tc": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-d",
            obs_vars=[
                "concNhno3",
                "concNtno3",
                "concNtnh",
                "concNnh3",
                "concnh4",
                "concpm10",
                "concpm25",
                "concSso2",
                "concNno2",
                "vmrco",
                "vmro3max",
                "vmro3",
                "concNno",
                "concso4t",
                "concso4c",
            ],
            obs_vert_type="Surface",
            colocate_time=True,
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            ts_type="daily",
            obs_filters=EBAS_FILTER,
        ),
        "EBAS-m-tc-ecoc": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-m",
            obs_vars=[
                "concCecpm25",
                "concCocpm25",
                "concom1",
                "concCecpm10",
                "concCocpm10",
                #                "concnh4pm10",
                "concnh4pm25",
                "concnh4pm1",
                #                "concso4pm10",
                "concso4pm25",
                "concso4pm1",
                "concno3pm10",
                "concno3pm25",
                "concno3pm1",
                "concsspm10",
                "concsspm25",
            ],
            obs_vert_type="Surface",
            colocate_time=True,
            ts_type="monthly",
            min_num_obs=OC_EC_RESAMPLE_CONSTRAINTS,
            obs_filters=EBAS_FILTER,
        ),
        "EBAS-d-tc-ecoc": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-d",
            obs_vars=[
                "concCecpm25",
                "concCocpm25",
                "concom1",
                "concCecpm10",
                "concCocpm10",
                # "concnh4pm10",
                "concnh4pm25",
                "concnh4pm1",
                #                "concso4pm10",
                "concso4pm25",
                "concso4pm1",
                "concno3pm10",
                "concno3pm25",
                "concno3pm1",
                "concsspm10",
                "concsspm25",
            ],
            obs_vert_type="Surface",
            colocate_time=True,
            ts_type="daily",
            min_num_obs=OC_EC_RESAMPLE_CONSTRAINTS,
            obs_filters=EBAS_FILTER,
        ),
        # Diurnal
        "EBAS-h-diurnal": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-h",
            obs_vars=[
                "concNno2",
                "concNno",
                "vmro3",
                "concpm10",
                "concpm25",
            ],
            obs_vert_type="Surface",
            ts_type="hourly",
            # diurnal_only=True,
            resample_how="mean",
            obs_filters={**EBAS_FILTER, "ts_type": "hourly"},
        ),
        # OX
        "EBAS-d-ox": dict(
            obs_id="EBAS-ox",
            obs_vars=["vmrox"],
            obs_type="ungridded",
            obs_vert_type="Surface",
            web_interface_name="EBAS",
            ts_type="daily",
            obs_merge_how={
                "vmrox": "eval",
            },
            obs_aux_requires={
                "vmrox": {
                    "EBASMC": [
                        "vmro3",
                        "vmrno2",
                    ],
                }
            },
            obs_aux_funs={
                "vmrox":
                # variables used in computation method need to be based on AeroCom
                # units, since the colocated StationData objects (from which the
                # new UngriddedData is computed, will perform AeroCom unit check
                # and conversion)
                "(EBASMC;vmro3+EBASMC;vmrno2)"
            },
            obs_aux_units={"vmrox": "nmol mol-1"},
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            obs_filters=EBAS_FILTER,
        ),
        "EBAS-h-ox-diurnal": dict(
            obs_id="EBAS-ox-diurnal",
            obs_vars=["vmrox"],
            obs_type="ungridded",
            obs_vert_type="Surface",
            web_interface_name="EBAS-h",
            ts_type="hourly",
            # diurnal_only=True,
            obs_merge_how={
                "vmrox": "eval",
            },
            obs_aux_requires={
                "vmrox": {
                    "EBASMC": ["vmro3", "vmrno2"],
                }
            },
            obs_aux_funs={
                "vmrox":
                # variables used in computation method need to be based on AeroCom
                # units, since the colocated StationData objects (from which the
                # new UngriddedData is computed, will perform AeroCom unit check
                # and conversion)
                "(EBASMC;vmro3+EBASMC;vmrno2)"
            },
            obs_aux_units={"vmrox": "nmol mol-1"},
            obs_filters={**EBAS_FILTER, "ts_type": "hourly"},
        ),
        # Wet Dep
        "EBAS-d-wet": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-d",
            ts_type="daily",
            obs_remove_outliers=True,
            obs_vars=[
                "wetoxs",
                "wetoxn",
                "wetrdn",
                "prmm",
            ],
            obs_vert_type="Surface",
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            colocate_time=True,
            obs_filters=EBAS_FILTER,
        ),
        "EBAS-m-wet": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-m",
            ts_type="monthly",
            obs_remove_outliers=True,
            colocate_time=True,
            obs_vars=[
                "wetoxs",
                "wetoxn",
                "wetrdn",
                "prmm",
            ],
            obs_vert_type="Surface",
            obs_filters=EBAS_FILTER,
        ),
        ################
        #    EEA-rural
        ################
        "EEA-d-rural": dict(
            obs_id="EEA-d-rural",
            obs_vars=[
                "concpm10",
                "concpm25",
                "concSso2",
                "concNno2",
                "concNno",
                "vmro3max",
            ],
            pyaro_config={
                "name": "EEA-d-rural",
                "reader_id": "eeareader",
                "filename_or_obj_or_url": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA-AQDS/download",
                "name_map": {
                    "PM2.5": "concpm25",
                    "PM10": "concpm10",
                    "NO": "concno",
                    "NO2": "concno2",
                    "SO2": "concso2",
                    "O3": "conco3",
                },
                "filters": {
                    "time_bounds": {
                        "startend_include": [
                            (f"{year}-01-01 00:00:00", f"{year + 1}-01-01 00:00:00")
                        ],
                    },
                    "valleyfloor_relaltitude": {
                        "topo": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/GTOPO30/merged",
                        "radius": 5000,
                        "topo_var": "Band1",
                        "lower": None,
                        "upper": 500,
                    },
                },
                "post_processing": [
                    "concNno_from_concno",
                    "concNno2_from_concno2",
                    "concSso2_from_concso2",
                    "vmro3max_from_conco3",
                ],
                "dataset": "verified",
                "station_area": [
                    "rural",
                    "rural-regional",
                    "rural-nearcity",
                    "rural-remote",
                ],
                "station_type": [
                    "background",
                ],
            },
            web_interface_name="EEA-rural",
            obs_vert_type="Surface",
            obs_filters=EEA_FILTER,
            ts_type="daily",
        ),
        "EEA-h-diurnal-rural": dict(
            obs_id="EEA-h-diurnal-rural",
            obs_vars=[
                "concNno2",
                "vmro3",
                "vmrox",
            ],
            pyaro_config={
                "name": "EEA-h-diurnal-rural",
                "reader_id": "eeareader",
                "filename_or_obj_or_url": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA-AQDS/download",
                "name_map": {
                    "PM2.5": "concpm25",
                    "PM10": "concpm10",
                    "NO": "concno",
                    "NO2": "concno2",
                    "SO2": "concso2",
                    "O3": "conco3",
                },
                "filters": {
                    "time_bounds": {
                        "startend_include": [
                            (f"{year}-01-01 00:00:00", f"{year + 1}-01-01 00:00:00")
                        ],
                    },
                    "valleyfloor_relaltitude": {
                        "topo": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/GTOPO30/merged",
                        "radius": 5000,
                        "topo_var": "Band1",
                        "lower": None,
                        "upper": 500,
                    },
                },
                "post_processing": [
                    "concNno2_from_concno2",
                    "vmro3_from_conco3",
                    "vmrno2_from_concno2",
                    "vmrox_from_vmrno2_vmro3",
                ],
                "dataset": "verified",
                "station_area": [
                    "rural",
                    "rural-regional",
                    "rural-nearcity",
                    "rural-remote",
                ],
                "station_type": [
                    "background",
                ],
            },
            web_interface_name="EEA-h-rural",
            obs_vert_type="Surface",
            obs_filters={**EEA_FILTER, "ts_type": "hourly"},
            resample_how="mean",
            ts_type="hourly",
        ),
        ################
        #    EEA-all
        ################
        "EEA-d-all": dict(
            obs_id="EEA-d-all",
            obs_vars=[
                "concpm10",
                "concpm25",
                "concSso2",
                "concNno2",
                "concNno",
                "vmro3max",
            ],
            pyaro_config={
                "name": "EEA-d-all",
                "reader_id": "eeareader",
                "filename_or_obj_or_url": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA-AQDS/download",
                "name_map": {
                    "PM2.5": "concpm25",
                    "PM10": "concpm10",
                    "NO": "concno",
                    "NO2": "concno2",
                    "SO2": "concso2",
                    "O3": "conco3",
                },
                "filters": {
                    "time_bounds": {
                        "startend_include": [
                            (f"{year}-01-01 00:00:00", f"{year + 1}-01-01 00:00:00")
                        ],
                    },
                    "valleyfloor_relaltitude": {
                        "topo": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/GTOPO30/merged",
                        "radius": 5000,
                        "topo_var": "Band1",
                        "lower": None,
                        "upper": 500,
                    },
                },
                "post_processing": [
                    "concNno_from_concno",
                    "concNno2_from_concno2",
                    "concSso2_from_concso2",
                    "vmro3max_from_conco3",
                ],
                "dataset": "verified",
            },
            web_interface_name="EEA-all",
            obs_vert_type="Surface",
            obs_filters=EEA_FILTER,
            ts_type="daily",
        ),
        "EEA-h-diurnal-all": dict(
            obs_id="EEA-h-diurnal-all",
            obs_vars=[
                "concNno2",
                "vmro3",
                "vmrox",
            ],
            pyaro_config={
                "name": "EEA-h-diurnal-all",
                "reader_id": "eeareader",
                "filename_or_obj_or_url": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/EEA-AQDS/download",
                "name_map": {
                    "PM2.5": "concpm25",
                    "PM10": "concpm10",
                    "NO": "concno",
                    "NO2": "concno2",
                    "SO2": "concso2",
                    "O3": "conco3",
                },
                "filters": {
                    "time_bounds": {
                        "startend_include": [
                            (f"{year}-01-01 00:00:00", f"{year + 1}-01-01 00:00:00")
                        ],
                    },
                    "valleyfloor_relaltitude": {
                        "topo": "/lustre/storeB/project/aerocom/aerocom1/AEROCOM_OBSDATA/GTOPO30/merged",
                        "radius": 5000,
                        "topo_var": "Band1",
                        "lower": None,
                        "upper": 500,
                    },
                },
                "post_processing": [
                    "concNno2_from_concno2",
                    "vmro3_from_conco3",
                    "vmrno2_from_concno2",
                    "vmrox_from_vmrno2_vmro3",
                ],
                "dataset": "verified",
            },
            web_interface_name="EEA-h-all",
            obs_vert_type="Surface",
            obs_filters={**EEA_FILTER, "ts_type": "hourly"},
            resample_how="mean",
            ts_type="hourly",
        ),
        ##################
        #    AERONET
        ##################
        "AERONET": dict(
            obs_id="AeronetSunV3Lev1.5.daily",
            obs_vars=["od550aer"],
            web_interface_name="AERONET",
            obs_vert_type="Column",
            ignore_station_names="DRAGON*",
            ts_type="daily",
            colocate_time=True,
            min_num_obs=dict(
                yearly=dict(
                    daily=90,
                ),
                monthly=dict(
                    weekly=1,
                ),
            ),
            obs_filters=AERONET_FILTER,
        ),
    }

    # Setup for supported satellite evaluations
    OBS_SAT = {}

    OBS_CFG = {**OBS_GROUNDBASED, **OBS_SAT}

    CFG["obs_cfg"] = OBS_CFG

    return copy.deepcopy(CFG)



import os.path
import pathlib

import pyaro
import aerovaldb
import aerovaldb.routes

from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
from pyaerocom.aeroval.config.ciconfigs.base_config import get_CFG
from pyaerocom.io import ReadUngridded
from pyaerocom.io import ReadGridded

import pytest

reportyear = year = 2018
CFG = get_CFG(
    reportyear=reportyear,
    year=year,
)

TEST_FILE = "mep-rd-Birkenes-2018-001.nc"


@pytest.mark.skip(
    reason="This test is skipped because the pyaro reader is not available in the current tox environment for testing on CI."
)
def test_harp_reader_available():
    """quick test to make sure pyaro_readers is working"""
    engine = "harp"
    assert engine in pyaro.list_timeseries_engines()


@pytest.mark.skip(
    reason="This test is skipped because the pyaro reader is not available in the current tox environment for testing on CI."
)
def test_harp_test_data_available():
    obs_config = CFG["obs_cfg"]
    conf_name = list(obs_config.keys())[0]
    data_path = pathlib.Path(
        dict(CFG["obs_cfg"][conf_name]["pyaro_config"])["filename_or_obj_or_url"]
    )
    tmp = data_path.glob("*.nc")
    data_files = [x.name for x in tmp if x.is_file()]
    assert len(data_files) > 0
    assert TEST_FILE in data_files


@pytest.mark.skip(
    reason="This test is skipped because the pyaro reader is not available in the current tox environment for testing on CI."
)
def test_obs_data_readable():
    """test reading obs data using pyaerocom"""

    reportyear = year = 2018
    CFG = get_CFG(
        reportyear=reportyear,
        year=year,
    )
    for aeroval_obs_name in CFG["obs_cfg"]:
        config = CFG["obs_cfg"][aeroval_obs_name]["pyaro_config"]
        reader = ReadUngridded(configs=config)
        data = reader.read()
        assert data


def test_aux_file_available():
    """test if the aux file is available"""

    reportyear = year = 2018
    CFG = get_CFG(
        reportyear=reportyear,
        year=year,
    )
    assert os.path.exists(CFG["io_aux_file"])


def test_model_data_readable():
    """test reading model data using pyaerocom"""

    reportyear = year = 2018
    CFG = get_CFG(
        reportyear=reportyear,
        year=year,
    )
    assert CFG
    for aeroval_model_name in CFG["model_cfg"]:
        #
        data = ReadGridded(data_id=CFG["model_cfg"][aeroval_model_name]["model_id"])
        assert data


@pytest.mark.skip(
    reason="This test is skipped because the pyaro reader is not available in the current tox environment for testing on CI."
)
def test_aeroval_config_diurnal():
    """test to make sure diurnal cycle analysis works
    The data used is entirely fake
    This test just checks if json files below ts/diurnal have been created"""

    reportyear = year = 2018
    CFG = get_CFG(
        reportyear=reportyear,
        year=year,
    )

    stp = EvalSetup(**CFG)
    ana = ExperimentProcessor(stp)

    ana.run()
    diurnals = ana.exp_output.avdb.query(aerovaldb.routes.Route.TIMESERIES_WEEKLY)
    assert len(diurnals) > 0



"""
High level I/O utility methods for pyaerocom
"""

from pyaerocom import change_verbosity, const
from pyaerocom.io.aerocom_browser import AerocomBrowser
from pyaerocom.io.readgridded import ReadGridded
from pyaerocom.io.readungridded import ReadUngridded


def get_ungridded_reader(obs_id):
    for reader in ReadUngridded.SUPPORTED_READERS:
        if obs_id in reader.SUPPORTED_DATASETS:
            return reader
    raise ValueError(f"No ungridded reader found that supports {obs_id}")


def browse_database(model_or_obs, verbose=False):
    """Browse Aerocom database using model or obs ID (or wildcard)

    Searches database for matches and prints information about all matches
    found (e.g. available variables, years, etc.)

    Parameters
    ----------
    model_or_obs : str
        model or obs ID or search pattern
    verbose : bool
        if True, verbosity level will be set to debug, else to critical

    Returns
    -------
    list
        list with data_ids of all matches

    Example
    -------
    >>> import pyaerocom as pya
    >>> pya.io.utils.browse_database('AATSR*ORAC*v4*')
    <BLANKLINE>
    Pyaerocom ReadGridded
    ---------------------
    Data ID: AATSR_ORAC_v4.02
    ...

    """
    if not verbose:
        change_verbosity("critical")
    else:
        change_verbosity("debug")
    browser = AerocomBrowser()
    matches = browser.find_matches(model_or_obs)
    if len(matches) == 0:
        print(f"No match could be found for {model_or_obs}")
        return
    elif len(matches) > 20:
        print(
            f"Found more than 20 matches for input pattern {model_or_obs}:\n\n"
            f"Matches: {matches}\n\n"
            f"To receive more detailed information, please specify search ID more accurately"
        )
        return
    for match in matches:
        try:
            if match in const.OBS_IDS_UNGRIDDED:
                reader = ReadUngridded(match)
            else:
                reader = ReadGridded(match)
            print(reader)
        except Exception as e:
            print(f"Reading failed for {match}. Error: {repr(e)}")
    return matches
